/* $Workfile: OsaLinux.cpp $   $Revision: 4 $    $Modtime: 4/02/13 12:07p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2013, All Rights Reserved

*******************************************************************************/

#include "../Osa.h"
#include "../OsaImplementation.h"
#include "../OsaRtl.h"
#include "../OsaMemDbg.h"
#include "Lib/OsaIoctl.h"
#include "Lib/OtxHwIds.h"
#include "OsaLinuxEntry.h"
#include "OsaLinuxCppRtl.h"


#define new ident_new
#define delete ident_delete
#define class ident_class
#define virtual ident_virtual			  
#define namespace ident_namespace
#define private ident_private
#define typename ident_typename

#ifndef CPP_ASMLINKAGE
#define CPP_ASMLINKAGE 
#else
#undef CPP_ASMLINKAGE
#define CPP_ASMLINKAGE 
#endif
#ifndef asmlinkage
#define asmlinkage CPP_ASMLINKAGE
#else
#undef asmlinkage
#define asmlinkage CPP_ASMLINKAGE
#endif

#undef new
#undef delete
#undef class
#undef virtual		
#undef namespace	
#undef private
#undef typename

#define ENABLE_WORKQUEUE_DBG 0 // 1=enable 0=not enabled


extern unsigned int nfnOsaLinuxCsIoControl;


/******************************************************************************
******************************************************************************/
enum OsaLinuxDeviceTypesE {
  OSA_LINUX_DEVICE_EXTERNAL = 0,
  OSA_LINUX_DEVICE_OS_EXT,
  OSA_LINUX_DEVICE_USER_EVENT
};

/******************************************************************************
******************************************************************************/
struct OsaLinuxBaseDevice {
  OsaLinuxBaseDevice()
    : m_pProcEntry(NULL)
    , m_eLinuxDeviceType(OSA_LINUX_DEVICE_EXTERNAL)
  {}

  proc_dir_entry_otx *m_pProcEntry;
  OsaLinuxDeviceTypesE m_eLinuxDeviceType;   // e.g. an OsaUserEventDevice
};


/******************************************************************************
  Note the similarity with the regular OsaEventImpl (OsaEvent), however
  this type of event inherits from the OsaLinuxBaseDevice and also features
  the PollWait member function so that it can be made accessible from the
  poll method (like an old select() method, almost)
******************************************************************************/
struct OsaLinuxUserEventDevice: public OsaLinuxBaseDevice {
  OsaLinuxUserEventDevice()
    : m_pWaitQHead(NULL)
    , m_nEventSet(0)
    , m_pClassValidityObject(&ms_classValidityObject)
  {
    OsaLinuxUserEventDevice_Ctor(&m_pWaitQHead);
  }

  ~OsaLinuxUserEventDevice()
  {
    SetEvent();
    OsaLinuxUserEventDevice_Dtor(&m_pWaitQHead);
    m_pClassValidityObject = NULL;
  }

  bool SetEvent()
  {
    OsaUint32 nEventSetBefore;

    OsaSingleLock localLock(&m_accessLock);
    nEventSetBefore = m_nEventSet;
    m_nEventSet = 1;
    OsaLinuxUserEventDevice_wake_up(&m_pWaitQHead);
    return nEventSetBefore ? true : false;
  }

  bool IsSet()
  {
    OsaSingleLock localLock(&m_accessLock);
    return m_nEventSet ? true : false;
  }

  void PollWait(void *pFile, void *pPollTable)
  {
    osalinux_poll_wait(pFile, &m_pWaitQHead, pPollTable);
  }

  void ClearEvent()
  {
    OsaSingleLock localLock(&m_accessLock);
    m_nEventSet = 0;
  }

  static int        ms_classValidityObject;
  void              *m_pClassValidityObject;

private:
  void              *m_pWaitQHead;
  OsaUint32         m_nEventSet;
  OsaSpinLock       m_accessLock;
};


/******************************************************************************
******************************************************************************/
struct OsaLinuxDevice: public OsaLinuxBaseDevice, public OsaDevice {
  OsaLinuxDevice()
    : m_pPciDev(NULL)
    , m_bInterruptConnected(false)
    , m_pDeviceCompanionDriver(NULL)
    , m_sTasklets1(NULL)
  {}

public:

  /******************************************************************************
    OsaLinuxAddToWorkQueue0
    Work queue function to add this object's work to the queue. 
    Need a unique function for each. The work queue will only execute
    one if two IRQ's are called before the queue can start the first one.
  ******************************************************************************/
  static void OsaLinuxAddToWorkQueue0(IN OsaULong queueData)
  {
    if(queueData == 0) {
      OSA_DPRINTF(("Nothing to do!! queueData is 0x%lx \n", 
                  queueData));
      return;
    }
    
    // Part of OsaLinuxEntry.c to finish the taskqueue startup and perform
    // cleanup as needed for the taskqueue. 
    OsaLinuxIsrWorkQueue(queueData);
  }
  /******************************************************************************
    OsaLinuxAddToWorkQueue1
  ******************************************************************************/
  static void OsaLinuxAddToWorkQueue1(IN OsaULong queueData)
  {
    if(queueData == 0) {
      OSA_DPRINTF(("Nothing to do!! queueData is 0x%lx \n", 
                  queueData));
      return;
    }
    
    // Part of OsaLinuxEntry.c to finish the taskqueue startup and perform
    // cleanup as needed for the taskqueue. 
    OsaLinuxIsrWorkQueue(queueData);
  }
  /******************************************************************************
    OsaLinuxAddToWorkQueue2
  ******************************************************************************/
  static void OsaLinuxAddToWorkQueue2(IN OsaULong queueData)
  {
    if(queueData == 0) {
      OSA_DPRINTF(("Nothing to do!! queueData is 0x%lx \n", 
                  queueData));
      return;
    }
    
    // Part of OsaLinuxEntry.c to finish the taskqueue startup and perform
    // cleanup as needed for the taskqueue. 
    OsaLinuxIsrWorkQueue(queueData);
  }
  /******************************************************************************
    OsaLinuxSetWorkQueuePtr
    Helper function to setup the workqueue pointer.
  ******************************************************************************/
  void OsaLinuxSetWorkQueuePtr(void)
  {
    int queueOk;

    m_nIsrIndex++;
    if(m_nIsrIndex >= 3) {
      m_nIsrIndex = 0;
    }

    // Set the workqueue function for this non-critical interrupt
    switch (m_nIsrIndex) {
      case 0:  
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue0); 
        break;      
      case 1:  
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue1); 
        break;      
      case 2:  
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue2); 
        break;      

      default:
        // This should never happen
        m_nIsrIndex = 0; // try and get it back in sync.
        m_pWorkqueueFunc = (void *) &(this->OsaLinuxAddToWorkQueue0);
        // This could cause an IRQ to be lost (not processed).
        OSA_EPRINTF(("OsaLinux: SoftIrq start function synchronization error.\n"));
    }

    // Pass a OsaLinuxDevice object pointer to the work queue
    queueOk = OsaLinuxAddWorkQueue((void *)&m_sTasklets1, 
                                   (void *)&(m_nWorkQueueData),
                                   (void *)&(*m_pWorkqueueFunc));
    if(queueOk != 0) {
      OSA_EPRINTF(("OsaLinux: Not able to add task to softIsr\n"));
    }


    return;
  }

  void   *m_pPciDev;
  bool    m_bInterruptConnected;
  void   *m_pDeviceCompanionDriver;  // A pointer to the implementation of this device in
                                     // companion driver (e.g.the PCMCLA client driver)

  OsaULong volatile m_nWorkQueueData;
  void   *m_nPendingInterrupts;
  bool    m_bSoftIsrActive;
  void * volatile m_sTasklets1;
  void (*m_pWorkqueueFunc)(IN OsaUint64);
  // Each card needs a unique workqueue starting task,
  // this index keep track of the last one used so each interrupt
  // will get a unique funciton. (If 2 interrupt functions are sent at the 
  // same time the workqueue only starts one of them.)
  OsaUint32 m_nIsrIndex;  
};


/******************************************************************************
******************************************************************************/
static void OsaLinuxInit();

static int OsaLinuxEnumAndAddAndStartAllPciBoards();

static bool OsaLinuxAddDevice(
  IN OsaBusTypeE eBusType,
  IN OsaUint16 nVid,
  IN OsaUint16 nDid,
  IN OsaUint16 nSid,
  IN wchar_t *pwcszVendorString,
  OUT OsaLinuxDevice **ppExternalDevice
  );

static bool OsaLinuxExtractPciResources(
  IN void *pPciDev,
  IN OsaResourcesS **ppResources
  );

static void OsaLinuxReturnResources(
  IN OsaLinuxDevice *pExternalDevice
  );


static bool OsaLinuxStartDevice(
  IN OsaLinuxDevice *pExternalDevice,
  IN OsaResourcesS *pResourceList
  );

static bool OsaLinuxRemoveDevice(
  IN OsaLinuxDevice *pExternalDevice
  );


static void OsaLinuxWorkQueueIsr(IN void *pContext);


static int ms_coEventsCreated = 0;
int OsaLinuxUserEventDevice::ms_classValidityObject;

/******************************************************************************
  Entry function exported to the pure "C" init_module()
******************************************************************************/
int OsaLinuxDriverEntry(void)
{
  int rc;
  OSA_DPRINTF(("OsaLinux: Creating OSA \n"));
  // Create Linux specific OSA

  OsaLinuxInit();

  // The drivers main entry point, must be exported as extern "C"
  // Sets up the other entry points in the form of pointers, very much like
  // OdOsa does for the operating system. This function is the only
  // function the driver must export so that Osa can bootstrap the driver into action.
  OsaString dummy;
#if 1 // OsaString does not work properly on empty strings, so we assign a dummy value to it /HJ 2000-10-18
  dummy.Append("Dummy");
#endif


  if (OSA_SUCCESS(::OsaDriverEntry(dummy.GetWcString()))) {

    if (!OsaCheckCallbacks()) {
      OSA_EPRINTF(("OsaLinux: The OSA base set of callbacks were not hooked up\n"));
      return -1;
    }

    // Enumerate OTX PCI boards
    rc = OsaLinuxEnumAndAddAndStartAllPciBoards();

    if (rc) {
      OSA_DPRINTF(("OsaLinux: Calling cleanup_module before failing the loading, do we need to do this?\n"));
      cleanup_module(); // Must find out if driver loads or not when returning an error code
    } else {
      // Create a bootstrap device for OS extensions (name to be known by user mode)
      OsaLinuxBaseDevice *pOsExt = new OsaLinuxBaseDevice;
      pOsExt->m_pProcEntry = OsaLinuxCreateProcDevice(L"osaext", pOsExt);
      pOsExt->m_eLinuxDeviceType = OSA_LINUX_DEVICE_OS_EXT;
    }

    return rc;
  } else {
    return -1;
  }

  return 0;
}


/******************************************************************************
  Entry function exported to the pure "C" cleanup_module()
******************************************************************************/
void OsaLinuxUnloadEntry()
{
  if (g_sOsaDriverEntryPoints.pfnUnload) {
    (*g_sOsaDriverEntryPoints.pfnUnload)();
  }

#if DBG
  if (OsaMemoryGetCurrentUsage()) {
    OSA_EPRINTF(("OsaLinux: Memory blocks no freed:\n"));
    OsaPrintCurrentlyUsed();
    OSA_EPRINTF(("OsaLinux: Dynamic memory leak is %d bytes\n", OsaMemoryGetCurrentUsage()));
  } else {
    OSA_DPRINTF(("OsaLinux: All dynamic memory freed properly\n"));
  }
#endif

  return;
}


/******************************************************************************
******************************************************************************/
void OsaLinuxInit()
{
  OsaInit();
  OsaInitLinuxRtl();
}


/******************************************************************************

  Called by the driver whenever the appearance of a named device must be
  presented to user mode apps or other drivers or the system or a combination
  of all. In particular, all board devices *must* be defined for the system, with
  a visible name, hence this function must be called from the drivers AddBoard
  function at the least.

  Od must supply a reference by which the device will be known in Od.

  The function returns a reference by which the device is known in Osa.

******************************************************************************/
OsaResultE OsaCreateExternalDevice(
  IN const OsaString &strDeviceName,            // The device name we want to present externally
  IN OsaInternalDeviceRep internalDeviceRep,       // A reference by which the device is known internally
  OUT OsaExternalDeviceRep *pExternalDeviceRep
)
{
  OsaLinuxDevice *pExternalDevice = new OsaLinuxDevice;
  OSA_ASSERT(pExternalDevice);
  if (!pExternalDevice) return OSA_RES_FAILED;

  OSA_DPRINTF(("OsaLinux: Created external device 0x%p\n", pExternalDevice));

  *pExternalDeviceRep = NULL;

  pExternalDevice->m_pProcEntry = OsaLinuxCreateProcDevice(strDeviceName.GetWcString(), pExternalDevice);
  pExternalDevice->m_eLinuxDeviceType = OSA_LINUX_DEVICE_EXTERNAL;

  if (!pExternalDevice->m_pProcEntry) return OSA_RES_FAILED;

  // Save the name, chosen by the internal part the driver, in the device extension
  pExternalDevice->m_strName = strDeviceName;

  pExternalDevice->m_internalDeviceRep = internalDeviceRep;

  *pExternalDeviceRep =  pExternalDevice;

  return OSA_RES_SUCCESS;
}


/******************************************************************************
******************************************************************************/
void OsaDestroyExternalDevice(
  IN OsaExternalDeviceRep externalDeviceRep       // A reference by which the device is known internally
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(externalDeviceRep);

  OSA_DPRINTF(("OsaLinux: Destroying external device 0x%p\n", pExternalDevice));
  OsaLinuxRemoveProcDevice(pExternalDevice->m_pProcEntry);
  pExternalDevice->m_pProcEntry = NULL;
  if(pExternalDevice->m_sTasklets1 != NULL) osalinux_kfree(pExternalDevice->m_sTasklets1);
  delete pExternalDevice;
}



/******************************************************************************
******************************************************************************/
bool OsaDoBoardSyncCall(IN OsaExternalDeviceRep pBoard, IN OsaSyncCallFnP pfnSyncCall, IN OUT void *pContext)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  OSA_ASSERT(pExternalDevice && pExternalDevice->m_bBoard && pfnSyncCall);

  if (!pExternalDevice || !pExternalDevice->m_bBoard || !pfnSyncCall) {
    return false;
  }

  OsaULong nFlags;
  osalinux_save_flags_cli(&nFlags);

  bool bResult = (*pfnSyncCall)(pExternalDevice->m_internalDeviceRep, pContext);

  osalinux_restore_flags(nFlags);

  return bResult;
}

/******************************************************************************
******************************************************************************/
void OsaRequestDeferredBoardNotification(IN OsaExternalDeviceRep pBoard, OsaIrpP pIrp, void *pContext)
{
#if ENABLE_WORKQUEUE_DBG == 1
  static OsaULong ntstPtr = 1;
#endif
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  OSA_ASSERT(pExternalDevice && pExternalDevice->m_bBoard);


#if 0 // Enable here to use Tasklet to process the reset of the interrupt in a task (DPC)
  {
    if(pExternalDevice->m_bSoftIsrActive == true) {
      // The last IRQ enabled NC-interrupts but is still not to
      // the end of OsaLinuxWorkQueueIsr(). May need to put a
      // thread sleep at the start of the function so the first
      // one can finish. Can not sleep here this is an interrupt.
#if ENABLE_WORKQUEUE_DBG == 1
      OSA_DPRINTF(("OsaLinux: Soft ISR is active\n"));
#endif
    }
    void *workQueue = reinterpret_cast<OsaLinuxDevice *>(pBoard);
    // The work queue will need to know what interrupts were set
    pExternalDevice->m_nPendingInterrupts = pContext;
    // The tasklet needs an unsigned long that contains the data. If
    // we passed the object directly it would need a unsigned long to
    // store the object pointer. But this is a function that is used by
    // all cards in the PC so each card need it's own variable to store
    // the object pointer.
    // Store object pointer in ULONG and pass that to the work queue.
    pExternalDevice->m_nWorkQueueData = (OsaULong) workQueue;

    // Set the IRQ start function for this interrupt.
    pExternalDevice->OsaLinuxSetWorkQueuePtr();

#if ENABLE_WORKQUEUE_DBG == 1
    // This check is only needed for debug. 
    // If the pointer which starts here, then which is passed 
    // to the workqueue and is finally returned to OsaLinuxWorkQueueIsr(),  
    // changes then two workqueues are active. This should only happen 
    // if two or more cards are in the PC.
    if(ntstPtr != (OsaULong) &(pExternalDevice->m_sTasklets1)) {
      ntstPtr = (OsaULong) &(pExternalDevice->m_sTasklets1);
      OSA_DPRINTF(("SoftIsr tstPtr-1 = 0x%lx\n", ntstPtr));
    }
#endif

  }

#else // Process everything in the interrupt
  (*pExternalDevice->m_pfnDeferredBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep, (OsaIrpP) pIrp, pContext);
#endif


  return;
}





/******************************************************************************
  A static wrapper callback.
******************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
void OsaLinuxInterruptCallback(
  IN int nIrq,
  IN void *pContext,
  struct pt_regs *pRegs
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
  (*pExternalDevice->m_pfnBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep);
}
#else  // 2.6.32
void OsaLinuxInterruptCallback(
  IN int nIrq,
  IN void *pContext
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
  (*pExternalDevice->m_pfnBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep);
}
#endif


/******************************************************************************
  Local Work queue call back function
    This call is the start of a new "kernel thread:process:task" that was 
    started by the kernel using the work queue implementation.
******************************************************************************/
static void OsaLinuxWorkQueueIsr(IN void *pContext)
{
#if ENABLE_WORKQUEUE_DBG == 1
  static OsaUint32 workThreadCnt = 0; // No more then one thread per card
  static OsaULong ntstPtr = 3;
  static OsaULong ntstPtr2 = 0;
#endif
  OsaUint64 *tempPtr;
  OsaUint64 pendingInterrupts;

  tempPtr = (OsaUint64 *) pContext;
//  OSA_DPRINTF(("OsaLinux: pContext = 0x%x\n", tempPtr));

  // The pointer to the OsaLinuxDevice was passed to the work queue
  // need to cast the void * to a OsaLinuxDevice *.
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(tempPtr);

#if ENABLE_WORKQUEUE_DBG == 1
  // Number of active thread at the same time
  workThreadCnt++;
  if(workThreadCnt > 1) {
    // There is at least one other active thread. This should only cause 
    // issues if the ISR is not able to have multiple threads running.
    OSA_DPRINTF(("OsaLinux: %d workqueue thread are active\n", workThreadCnt));
  }

  // This check is only needed for debug. 
  // If the pointer set in OsaRequestDeferredBoardNotification()
  // which is passed to the workqueue has changed then two tasklets  
  // are active. This should only happen if two or more cards are in the PC.
  if(ntstPtr != (OsaULong) &(pExternalDevice->m_sTasklets1)) {
    ntstPtr = (OsaULong) &(pExternalDevice->m_sTasklets1);
//    OSA_DPRINTF(("SoftIsr tstPtr-3 = 0x%x\n", ntstPtr));
  }  
#endif

  if(pContext == 0) {
    OSA_DPRINTF(("OsaLinux: pContext = 0\n"));
  }
  else
  if(pExternalDevice == NULL) {
    OSA_DPRINTF(("OsaLinux: pExternalDevice = 0\n"));
  }
  else
  if((*pExternalDevice->m_pfnBoardNotificationCallback) == NULL) {
     OSA_DPRINTF(("OsaLinux: *pExternalDevice = 0\n"));
  }
  else {
#if ENABLE_WORKQUEUE_DBG == 1
    pExternalDevice->m_bSoftIsrActive = true; // Soft ISR is running
#endif
    pendingInterrupts = pExternalDevice->m_nPendingInterrupts;
    // This will call the objects interrupt service routine
    (*pExternalDevice->m_pfnDeferredBoardNotificationCallback)(pExternalDevice->m_internalDeviceRep, NULL, pendingInterrupts);
#if ENABLE_WORKQUEUE_DBG == 1
    // Note at this point a new DPC(thread) could start.
    // The NC-interrupts have been enabled.
    pExternalDevice->m_bSoftIsrActive = false; // Soft ISR is finished
#endif
  }

  // Finished this task. 
  // Note at this point a new DPC(thread) could start.
  // The NC-interrupts have been enabled.
#if ENABLE_WORKQUEUE_DBG == 1
  workThreadCnt--;
#endif

  return;
}

/******************************************************************************
  Work queue call back function. This function is called from OsaLinuxEntry.c
  It is the entry point for the interrupt object to the card.
******************************************************************************/
void OsaLinuxWorkQueueFunction(void *pContext)
{
  OsaLinuxWorkQueueIsr(pContext);
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
extern "C"
int OsaLinuxOpenDeviceEntry(void *pFile, void *pContext)
{
  OsaResultE eResult = OSA_RES_SUCCESS;
  OsaIrp sIrp;

  if (!pFile || !pContext) {
    OSA_DPRINTF(("OsaLinux: OsaLinuxOpenDeviceEntry: pFile:%p  pContext:%p\n", pFile, pContext));
    return -1;
  }

  sIrp.nIoCode          = 0;
  sIrp.pInputBuffer     = NULL;
  sIrp.pOutputBuffer    = NULL;
  sIrp.cbInput          = 0;
  sIrp.cbOutputCapacity = 0;
  sIrp.nOwnerId         = (OsaUint64) pFile;
  sIrp.cbOutput         = 0;

  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);

  switch (pBaseDevice->m_eLinuxDeviceType) {
  case OSA_LINUX_DEVICE_OS_EXT:
    OSA_DPRINTF(("OsaLinux: Opening OsExtension device\n"));
    break;
  case OSA_LINUX_DEVICE_USER_EVENT:
    {
      OSA_DPRINTF(("OsaLinux: Opening UserEvent device\n"));
    }
    break;
  default:
    {
      OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
      OSA_DPRINTF(("OsaLinux: Opening external device\n"));
      eResult = (*g_sOsaDriverEntryPoints.pfnOpenDevice)(pExternalDevice->m_internalDeviceRep, &sIrp);
    }
    break;
  }

  // Translate to Linux return codes
  switch (eResult) {
  case OSA_RES_SUCCESS:
    //OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning %d bytes\n", pHdr->cbReturned));
    return 0;
  default:
    OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning error (0) bytes\n"));
    return -1;
  }
}



/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
extern "C"
int OsaLinuxCloseDeviceEntry(void *pFile, void *pContext)
{
  OsaResultE eResult = OSA_RES_SUCCESS;
  OsaIrp sIrp;

  sIrp.nIoCode          = 0;
  sIrp.pInputBuffer     = NULL;
  sIrp.pOutputBuffer    = NULL;
  sIrp.cbInput          = 0;
  sIrp.cbOutputCapacity = 0;
  sIrp.nOwnerId         = (OsaUint64) pFile;
  sIrp.cbOutput         = 0;

  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);

  switch (pBaseDevice->m_eLinuxDeviceType) {
  case OSA_LINUX_DEVICE_OS_EXT:
    {
      // No special device class is used but we do the reinterpret_cast anyway (for consistency)
      OsaLinuxBaseDevice *pDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pBaseDevice);
      OSA_DPRINTF(("OsaLinux: Closing OsExtension device\n"));
      // The device is not deleted.
    }
    break;
  case OSA_LINUX_DEVICE_USER_EVENT:
    {
      OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
      OSA_DPRINTF(("OsaLinux: Closing User Event, proc_entry deleted\n"));
      OsaLinuxRemoveProcDeviceDeferred(pDevice->m_pProcEntry);
      delete pDevice;
    }
    break;
  default:
    {
      OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
      OSA_DPRINTF(("OsaLinux: Closing external device\n"));
      eResult = (*g_sOsaDriverEntryPoints.pfnCloseDevice)(pExternalDevice->m_internalDeviceRep, &sIrp);
    }
    break;
  }

  // Translate to Linux return codes
  switch (eResult) {
  case OSA_RES_SUCCESS:
    //OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning %d bytes\n", pHdr->cbReturned));
    return 0;
  default:
    OSA_DPRINTF(("OsaLinux: pfnCloseDevice returning error (0) bytes\n"));
    return -1;
  }
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
int OsaLinuxIoDeviceControlEntry(void *pFile, void *pContext, OsaUint32 nIoControlCode, OsaLinuxIoCtlHdr *pHdr, void *pSystemBuffer)
{
  OsaResultE eResult = OSA_RES_SUCCESS;
  OsaIrp sIrp;

  sIrp.nIoCode          = nIoControlCode;
  sIrp.pInputBuffer     = pSystemBuffer;
  sIrp.pOutputBuffer    = pSystemBuffer;
  sIrp.cbInput          = pHdr->cbInBufSize;
  sIrp.cbOutputCapacity = pHdr->cbOutBufSize;
  sIrp.nOwnerId         = (OsaUint64) pFile;
  sIrp.cbOutput         = 0;  // If the callee wants to return something he must set this value

  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);
  switch (pBaseDevice->m_eLinuxDeviceType) {

  case OSA_LINUX_DEVICE_OS_EXT:
    {
      // No special device class is used but we do the reinterpret_cast anyway (for consistency)
      OsaLinuxBaseDevice *pDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pBaseDevice);
      switch (nIoControlCode) {

      case OSA_LINUX_OSEXT_IO_CREATE_EVENT:
        {
          // Create a proc entry and a OsaLinuxUserEventDevice. Attach the two to eachouther and return
          // the chosen name of the event device to the user lib
          OsaLinuxIoOutCreateEvent *pOut = reinterpret_cast<OsaLinuxIoOutCreateEvent *>(sIrp.pOutputBuffer);

          OsaString strName;
          strName.Append(OSA_USER_EVENT_NAME_PREFIX);
          strName.Append(ms_coEventsCreated++);

          OsaLinuxUserEventDevice *pUserEvent = new OsaLinuxUserEventDevice;
          pUserEvent->m_pProcEntry = OsaLinuxCreateProcDevice(strName.GetWcString(), pUserEvent);
          pUserEvent->m_eLinuxDeviceType = OSA_LINUX_DEVICE_USER_EVENT;

          // Return the name so that the user mode lib can open it and get a user mode handle
          if (sIrp.cbOutputCapacity - sizeof(void*) > strName.GetLength()) {
            pOut->pEventDevice = pUserEvent;
            wcsztosz(pOut->szProcEntryName, strName.GetWcString());
            sIrp.cbOutput = sizeof(OsaLinuxIoOutCreateEvent);
          } else {
            eResult = OSA_RES_FAILED;
          }
        }
        break;

      case OSA_LINUX_OSEXT_IO_DELETE_EVENT:
        {
          OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
          OsaLinuxRemoveProcDevice(pDevice->m_pProcEntry);
          delete pDevice;
        }
        break;

      default:
        break;
      }

    }
    break;

  case OSA_LINUX_DEVICE_USER_EVENT:
    {
      OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
      switch (nIoControlCode) {
      case OSA_LINUX_USREV_IO_SET_EVENT:
        {
          pDevice->SetEvent();
        }
        break;

      default:
        OSA_DPRINTF(("OsaLinux: Unknown OsaLinux Internal I/O code 0x%x", nIoControlCode));
        break;
      }
    }
    break;
  default:
    OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);

    eResult = (*g_sOsaDriverEntryPoints.pfnIoControlDevice)(pExternalDevice->m_internalDeviceRep, &sIrp);
    break;
  }
  // Translate to NTWDM status codes
  switch (eResult) {
  case OSA_RES_SUCCESS:
    pHdr->cbReturned = sIrp.cbOutput;
    //OSA_DPRINTF(("OsaLinux: OsaLinuxIoDeviceControlEntry returning %d bytes\n", pHdr->cbReturned));
    return 0;
  default:
    OSA_EPRINTF(("OsaLinux: OsaLinuxIoDeviceControlEntry returning error (Info = 0x%X)\n", sIrp.cbOutput));
    pHdr->cbReturned = 0;
    return -1;
  }
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
extern "C"
unsigned int OsaLinuxPollDeviceEntry(void *pFile, void *pContext, void *pPollTable)
{
  OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);
  if (pBaseDevice && pBaseDevice->m_eLinuxDeviceType == OSA_LINUX_DEVICE_USER_EVENT) {
    OsaLinuxUserEventDevice *pDevice = reinterpret_cast<OsaLinuxUserEventDevice *>(pBaseDevice);
    pDevice->PollWait(pFile, pPollTable);
    if (pDevice->IsSet()) {
      return 1;
    }
  }
  return 0;
}


/******************************************************************************
  Entry function exported to the pure "C" OsaDriverEntry.c
******************************************************************************/
void OsaLinuxRemoveDeviceEntry(void *pContext)
{

  OSA_ASSERT(pContext);
  if (pContext) {

    OsaLinuxBaseDevice *pBaseDevice = reinterpret_cast<OsaLinuxBaseDevice *>(pContext);
    switch (pBaseDevice->m_eLinuxDeviceType) {
    case OSA_LINUX_DEVICE_OS_EXT:
    case OSA_LINUX_DEVICE_USER_EVENT:
      {
        OsaLinuxRemoveProcDevice(pBaseDevice->m_pProcEntry);
        delete pBaseDevice;
      }
      break;
    case OSA_LINUX_DEVICE_EXTERNAL:
      {
        OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pContext);
        OsaLinuxRemoveDevice(pExternalDevice);
      }
      break;
    default:
      OSA_DPRINTF(("OsaLinux: Unknown OsaLinux device type encountered, cannot remove\n"));

      break;
    }
  }
}

/******************************************************************************
******************************************************************************/
void OsaDisconnectNotificationCallbacks(
  IN OsaExternalDeviceRep externalDeviceRep
  )
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(externalDeviceRep);
  if (pExternalDevice->m_pResources) {

    // Disconnect the interrupt first if connected
    if (pExternalDevice->m_bInterruptConnected) {
      osalinux_free_irq(pExternalDevice->m_pResources->sInterrupt.nLevel, pExternalDevice);
      pExternalDevice->m_bInterruptConnected = false;
      OSA_DPRINTF(("OsaLinux: Interrupt %d Disconnected\n", pExternalDevice->m_pResources->sInterrupt.nLevel));
    } else {
      OSA_DPRINTF(("OsaLinux: No interrupt to disconnect\n"));
    }
  }
}

/******************************************************************************
******************************************************************************/
bool OsaConnectNotificationCallbacks(
  IN OsaExternalDeviceRep              externalDeviceRep,
  IN OsaBoardNotificationFnP           pfnBoardNotificationCallback,
  IN OsaDeferredBoardNotificationFnP   pfnDeferredBoardNotificationCallback,
  IN OsaResourcesS::InterruptResourceS *pInterruptResource
  )
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(externalDeviceRep);

  pExternalDevice->m_pfnDeferredBoardNotificationCallback = pfnDeferredBoardNotificationCallback;
  pExternalDevice->m_pfnBoardNotificationCallback = pfnBoardNotificationCallback;

  // PCI
  OSA_DPRINTF(("OsaLinux: Driver was started.\n"));

  if (!pExternalDevice->m_pResources->sInterrupt.bHaveInterrupt) {
    OSA_DPRINTF(("OsaLinux: !pExternalDevice->m_pResources->sInterrupt.bHaveInterrupt!\n"));
    return false;
  }

  int nError = osalinux_request_irq(pExternalDevice->m_pResources->sInterrupt.nLevel, 
                                      pExternalDevice);

  if (!nError) {
    OSA_DPRINTF(("OsaLinux: Interrupt %d connected\n", pExternalDevice->m_pResources->sInterrupt.nLevel));
    pExternalDevice->m_bInterruptConnected = true;
  } else {
    OSA_EPRINTF(("OsaLinux: Unexpected request_irq() error\n"));
    return false;
  }

  return true;
}


/******************************************************************************
******************************************************************************/
int OsaLinuxEnumAndAddAndStartAllPciBoards()
{
  int status = 0;  // '0' = Success

  OSA_DPRINTF(("OsaLinux: Enumerating PCI devices\n"));

#define CONFIG_PCI
#ifdef CONFIG_PCI
  // Note: All unix like identifiers (non-mixed case and underscore) are from the Linux kernel headers
  if (OsaLinux_pci_present()) {

    void *pPciDev;
    for (pPciDev = OsaLinux_pci_dev_next(NULL); pPciDev != NULL; pPciDev = OsaLinux_pci_dev_next(pPciDev)) {

      if(pPciDev == NULL) {
        OSA_DPRINTF(("OsaLinux: pPciDev NULL - Search Finished\n\n"));

        return(status);
      }
      // At this point we have found a board

      // Declare a functional device object (as opposed to a physical device object which is
      // present lower in the bus driver in WDM.
      // The object is allocated by Osa on request by Od
      OsaLinuxDevice *pExternalDevice = NULL;
      OsaInternalDeviceRep internalDeviceRep = NULL;

      OsaUint16 SubSystemID;
      OsaLinux_pci_read_config_word(pPciDev, PCI_SUBSYSTEM_ID, &SubSystemID);

      if (!OsaLinuxAddDevice(
        OSA_BUS_PCI,
        OsaLinux_pci_dev_vendor(pPciDev),
        OsaLinux_pci_dev_device(pPciDev),
        SubSystemID,
        NULL,
        OUT &pExternalDevice
        ))
      {
        // Probably not a device the Driver wanted to handle
       OSA_DPRINTF(("OsaLinux: Not a device the driver supports\n"));

       continue;
      }

      OSA_DPRINTF(("OsaLinux: Found and added PCI device\n"));

      OsaResourcesS *pOsaResources = NULL;

      pExternalDevice->m_pPciDev = pPciDev;

      OsaLinux_pci_dev_enable(pPciDev);
      // Claim resources, OsaLinuxRemoveDevice frees the resources (by calling OsaLinuxReturnResources)
      if (!OsaLinuxExtractPciResources(
        pPciDev,
        OUT &pOsaResources      // If the board is requesting resources we get the requests here (allocated with new)
        ))
      {
        // If error code we assume the AssignResources did not allocate pOsaResources for us
        // If error we should not get any resources allocated for us so we don't need to delete them
        OSA_EPRINTF(("OsaLinux: Failed to extract resources\n"));
        OSA_ASSERT(!pOsaResources);
        OsaLinuxRemoveDevice(pExternalDevice);
        continue;
      }

      OSA_DPRINTF(("OsaLinux: Extracted resources\n"));

      OSA_ASSERT(pOsaResources);  // All our boards request resources (how would the be used otherwise?)

      bool bResult = OsaLinuxStartDevice(pExternalDevice, pOsaResources);

      delete pOsaResources; // These are only the unmapped and unclaimed resources

      if (!bResult) {
        // If the device could not be started we must delete it
        // Any resources claimed will be released by OsaLinuxRemoveDevice calling OsaLinuxReturnResources
        OSA_DPRINTF(("<1>Failed to start OTX PCI device\n"));
        OsaLinuxRemoveDevice(pExternalDevice);
        status = -1;
        continue;
      } else {
        OSA_DPRINTF(("OsaLinux: Started device\n"));
      }

    }

  } else {
    OSA_DPRINTF(("<1>No PCI support in kernel, loading aborted\n"));
    return -1;
  }

  return status;
#else
  OSA_DPRINTF(("<1>No PCI support in kernel, loading aborted\n"));
  return -1;
#endif
}


/******************************************************************************
******************************************************************************/
bool OsaLinuxAddDevice(
  IN OsaBusTypeE  eBusType,
  IN OsaUint16 nVid,
  IN OsaUint16 nDid,
  IN OsaUint16 nSid,
  IN wchar_t *pwcszVendorString,
  OUT OsaLinuxDevice **ppExternalDevice
)
{
  *ppExternalDevice = NULL;
  OsaString strExternalDeviceName;
  OsaInternalDeviceRep internalDeviceRep = NULL;
  OsaLinuxDevice *pExternalDevice = NULL;

  // We now support both PCI
  if (!(eBusType == OSA_BUS_PCI )) {
    OSA_EPRINTF(("OsaLinux: Unsupported Bus Type %d\n", eBusType));
    return false;
  }

  if((nDid & (OsaUint16) 0xF000) == (OsaUint16) 0xF000) {
    // Thor-2-ExpressCard Bridge use F432 as its device ID. This
    // is not a valid device for the OTX driver. Use the 0xF000  
    // mask to determine a valid device ID. No Odin Device ID  
    // contains an 'F' in the MSB. Found in driver release 2.18.1.0.
    OSA_DPRINTF(("OsaLinux: Unsupported Device ID %d(0x%X) on Bus Type %d \n",
                 nDid, 
                 nDid, 
                 eBusType));
    return false;
  }

  if((nSid & (OsaUint16) 0xFFFF) == (OsaUint16) 0xFFFF) {
    // Thor-2-ExpressCard, if card is installed after the laptop
    // PC has been booted the Sub System ID is set to 0xFFFF. If
    // this is the case then starting the driver would cause the 
    // PC to crash because the card is not initialized correctly.
    // Found in driver release 2.18.1.0.
    OSA_EPRINTF(("OsaLinux: Unsupported SubSystem ID %d(0x%X) "
                 "Device ID 0x%X Bus Type %d \n",
                 nSid, 
                 nSid, 
                 nDid,
                 eBusType));
    return false;
  }

  if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnAddBoard)(eBusType, nVid, nDid, nSid, pwcszVendorString, &internalDeviceRep, &strExternalDeviceName))) {
    return false;
  }

  if (!OSA_SUCCESS(OsaCreateExternalDevice(
    strExternalDeviceName,                // Externally known name of the board device
    internalDeviceRep,                    // A reference by which the device is known internally (Driver)
    (void **) &pExternalDevice            // A reference by which the device is known externally (OSA and System)
    )))
  {
    // Cleanup after (*AddBoardFnP)() since if OsaLinuxAddDevice fails
    // the caller won't call OsaLinuxRemoveDevice (caller thinks device was never added)
    OSA_EPRINTF(("OsaLinux: Failed to create external device\n", eBusType));
    (*g_sOsaDriverEntryPoints.pfnRemoveBoard)(internalDeviceRep);
    return false;
  }

  OSA_ASSERT(pExternalDevice);
  pExternalDevice->m_bBoard = true;

  *ppExternalDevice = pExternalDevice;

  return true;
}


/******************************************************************************

  Figures out the resources needed for the device

  Cleanup needed: Caller is responsible for deleting *ppResources if return
  value is true

******************************************************************************/
bool OsaLinuxExtractPciResources(
  IN void *pPciDev,
  OUT OsaResourcesS **ppResources
)
{
  OSA_ASSERT(ppResources && !*ppResources);
  OSA_ASSERT(pPciDev);

  // If we return failure we must delete this resources object
  // If we return true we must attach it to the OUT parameter
  OsaResourcesS *pResources = new OsaResourcesS;

  OSA_ASSERT(pResources);

  if (!pResources) {
    return false;
  }


  // Determine what resources to use

  if (OsaLinux_pci_dev_irq(pPciDev)) {
    pResources->sInterrupt.bHaveInterrupt = true;
    pResources->sInterrupt.nLevel = OsaLinux_pci_dev_irq(pPciDev);
    // The interrupt is connected in ConnectNotificationCallbacks()
  }

  pResources->coPortAreas = 0;
  pResources->coMemoryAreas = 0;

  for (int idx = 0; idx <= 5; idx++) {
    u32 nPciBaseAddrRegOffset = PCI_BASE_ADDRESS_0 + 4*idx;

    // Figure out how much space and of what type this
    // device wants.
    //
    if (!OsaLinux_pci_dev_resource_start(pPciDev, idx)) {
      // this base-address register is unused
      // We perhaps should break the loop here since there should not be any more regions according to PCI specs
      continue;
    }

    u32 nPciOrigAddrVal;
    u32 nPciAddr;
    u32 mask;

    // During the following temporary remap of the PCI registers and ports we
    // should prevent any other kernel component from touching PCI memory
    // in a UP architecture this is easy, just disabling the interrups
    // and this 'thread' of execution owns the CPU. On an SMP archtecture
    // I don't know how to make this 'thread' king, now when the kernel
    // is using finer grained spinlocks instead of a big one.
    // The bios32.c kernel file uses a pci_lock, but that only seems
    // to protect config space.
    // I do an global_cli to prevent interrupts on all processors (I think)
    // and then I cross my fingers that no other processor is running an
    // dispatch thread doing PCI access.
    //

    long unsigned f;

    osalinux_save_flags_cli(&f);

    OsaLinux_pci_read_config_dword(pPciDev, nPciBaseAddrRegOffset, &nPciOrigAddrVal);
    OsaLinux_pci_write_config_dword(pPciDev, nPciBaseAddrRegOffset, 0xffffffff);
    OsaLinux_pci_read_config_dword(pPciDev, nPciBaseAddrRegOffset, &nPciAddr);

    // Write back the original value to the PCI board
    OsaLinux_pci_write_config_dword(pPciDev, nPciBaseAddrRegOffset, nPciOrigAddrVal);
    osalinux_restore_flags(f);

    // We've read the base address register back after
    // writing all ones and so now we must decode it.
    //
    if (nPciAddr & PCI_BASE_ADDRESS_SPACE_IO) {

      // I/O space base address register.
      //
      nPciAddr &= PCI_BASE_ADDRESS_IO_MASK;
      mask = (~nPciAddr << 1) | 0x1;

      pResources->sqIoPortArea[pResources->coPortAreas].cbLength = (mask & nPciAddr) & 0xffffffff;

      // Use the value the kernel has decided for us, not from the PCI device
      pResources->sqIoPortArea[pResources->coPortAreas].pStart.sWord.nLow = OsaLinux_pci_dev_resource_start(pPciDev, idx);

      pResources->coPortAreas++;
    } else {

      // Memory space base address register.
      //
      nPciAddr &= PCI_BASE_ADDRESS_MEM_MASK;
      mask = (~nPciAddr << 1) | 0x1;

      pResources->sqMemoryRegArea[pResources->coMemoryAreas].cbLength  = (mask & nPciAddr) & 0xffffffff;

      pResources->sqMemoryRegArea[pResources->coMemoryAreas].pStart.sWord.nLow = OsaLinux_pci_dev_resource_start(pPciDev, idx) & PCI_BASE_ADDRESS_MEM_MASK;  //yh: it was  PCI_BASE_ADDRESS_IO_MASK

      pResources->coMemoryAreas++;
    }

  }

  *ppResources = pResources;

  return true;
}


/******************************************************************************
  Starts a functional device object that represents a device on a bus (a board).

  This function is deliberately made similar to IRP_MN_START_DEVICE. However
  the resource list is passed as a distinct parameter an not in the stack of
  an IRP.

  If this function fails the caller must call OsaLinuxRemoveDevice to:
  Cleanup after (*AddBoardFnP)()
  Cleanup after partially mapped and claimed resources
******************************************************************************/
bool OsaLinuxStartDevice(
  IN OsaLinuxDevice *pExternalDevice,
  IN OsaResourcesS *pResources
)
{
  // Create a resource list for the claimed and mapped resources.
  OsaResourcesS *pClaimedResources = new OsaResourcesS;

  OSA_ASSERT(pClaimedResources);

  if (!pClaimedResources) {
    return false;
  }

  // We must attach the resources to the device so they get released with the device
  // We will soon start claiming resources and may return with an error
  // Any error returned by this function will result in a call to OsaLinuxRemoveDevice and OsaLinuxReturnResources
  // Return resources will return any resources we have claimed and mapped here, but only
  // if they can be found through pExternalDevice->m_pResources
  pExternalDevice->m_pResources = pClaimedResources;  // We have no resources claimed yet, but an empty OsaResourcesS object is okay too

  pClaimedResources->sInterrupt.bHaveInterrupt = pResources->sInterrupt.bHaveInterrupt;
  pClaimedResources->sInterrupt.nLevel = pResources->sInterrupt.nLevel;
  pClaimedResources->coPortAreas = 0;
  // Claim the I/O ports
  for (int iPort = 0; iPort < pResources->coPortAreas; iPort++) {
    OSA_DPRINTF(("OsaLinux: PCI I/O area length = %d\n", pResources->sqIoPortArea[iPort].cbLength));
    OSA_DPRINTF(("OsaLinux: PCI I/O area start (raw)= 0x%x\n", (OsaUint32) pResources->sqIoPortArea[iPort].pStart));

    if (osalinux_check_region(pResources->sqIoPortArea[iPort].pStart, pResources->sqIoPortArea[iPort].cbLength)) {
      // Region taken, since it's PCI which was setup by the kernel/bios this should not happen
      OSA_EPRINTF(("OsaLinux: Unexpected check_region() error\n"));
      return false;
    }

    // Always succeeeds if the check_region() above succeeded
    osalinux_request_region(pResources->sqIoPortArea[iPort].pStart, pResources->sqIoPortArea[iPort].cbLength, "OSA_Device");

    pClaimedResources->sqIoPortArea[iPort].pStart = pResources->sqIoPortArea[iPort].pStart;
    pClaimedResources->sqIoPortArea[iPort].cbLength = pResources->sqIoPortArea[iPort].cbLength;
    pClaimedResources->coPortAreas = iPort + 1;

    OSA_DPRINTF(("OsaLinux: PCI I/O area start (claimed)= 0x%x\n", (OsaUint32) pClaimedResources->sqIoPortArea[iPort].pStart));

  }

  // Map the physical memory addresses to virtual addresses
  pClaimedResources->coMemoryAreas = 0;
  for (int iMemory = 0; iMemory < pResources->coMemoryAreas; iMemory++) {

    OSA_DPRINTF(("OsaLinux: PCI Memory area length = %d\n", pResources->sqMemoryRegArea[iMemory].cbLength));
    OSA_DPRINTF(("OsaLinux: PCI Memory area start (unmapped)= 0x%x\n", (OsaUint32) pResources->sqMemoryRegArea[iMemory].pStart));


    pClaimedResources->sqMemoryRegArea[iMemory].pStart = osalinux_ioremap(pResources->sqMemoryRegArea[iMemory].pStart, pResources->sqMemoryRegArea[iMemory].cbLength);
    pClaimedResources->sqMemoryRegArea[iMemory].cbLength = pResources->sqMemoryRegArea[iMemory].cbLength;
    pClaimedResources->coMemoryAreas = iMemory + 1;
    if (pClaimedResources->sqMemoryRegArea[iMemory].pStart) {
      pClaimedResources->sqMemoryRegArea[iMemory].eAddressSpace = OsaResourcesS::MemorySpaceMapped;
      OSA_DPRINTF(("OsaLinux: PCI Memory area start (mapped)= 0x%x\n", (OsaUint32) pClaimedResources->sqMemoryRegArea[iMemory].pStart));
    } else {
      // Mapping failed
      pClaimedResources->sqMemoryRegArea[iMemory].cbLength = 0;
      OSA_EPRINTF(("OsaLinux: Unexpected map error\n"));
      return false;
    }
  }

  OSA_DPRINTF(("OsaLinux: %d I/O areas requested\n", pClaimedResources->coPortAreas));
  OSA_DPRINTF(("OsaLinux: %d Mem areas mapped\n", pClaimedResources->coMemoryAreas));

  // Start the board using the claimed and mapped resources.
  if (!OSA_SUCCESS((*g_sOsaDriverEntryPoints.pfnStartBoard)(pExternalDevice->m_internalDeviceRep, pExternalDevice, pExternalDevice->m_pResources))) {
    return false;
  }


  return true;
}


/******************************************************************************
  Removes a functional device object that represents a device on a bus (a board).

  This function is deliberately made similar to the OsaWdmRemoveDevice().

******************************************************************************/
bool OsaLinuxRemoveDevice(
  IN OsaLinuxDevice *pExternalDevice
)
{
  OSA_DPRINTF(("OsaLinux: Removing device\n"));
  // When removing in a Linux unload routine we have no choice, so the driver must be ready
  OsaResultE result = (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_CAN_REMOVE);
  OSA_ASSERT(OSA_SUCCESS(result));

  (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_REJECT_NEW_IRPS);
  (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_WAIT_FOR_IRPS_COMPLETED);
  (*g_sOsaDriverEntryPoints.pfnBoardState)(pExternalDevice->m_internalDeviceRep, OSA_BOARD_DISABLE_IO);

  (*g_sOsaDriverEntryPoints.pfnRemoveBoard)(pExternalDevice->m_internalDeviceRep);

  OsaLinuxReturnResources(pExternalDevice);

  OsaDestroyExternalDevice(pExternalDevice);

  return true;
}



/******************************************************************************
  This is function is called to return any hardware resources which
  have been allocated for the device.  In our case, these are
  the Interrupt Object, and any ports we might have mapped through
  memory, and all assigned resources.  Note that we do NOT return our
  Device Object here.
******************************************************************************/
void OsaLinuxReturnResources(
  OsaLinuxDevice *pExternalDevice
)
{
  OSA_DPRINTF(("OsaLinux: Returning Resources\n"));
  if (pExternalDevice->m_pResources) {
 
    // PCI

    OSA_DPRINTF(("OsaLinux: %d I/O areas to release\n", pExternalDevice->m_pResources->coPortAreas));
    OSA_DPRINTF(("OsaLinux: %d Mem areas to unmap\n", pExternalDevice->m_pResources->coMemoryAreas));

    // Now unmap memory address resources and release io ports
    for (int iPort = 0; iPort < pExternalDevice->m_pResources->coPortAreas; iPort++) {
      OSA_DPRINTF(("OsaLinux: Releasing I/O region 0x%p\n", pExternalDevice->m_pResources->sqIoPortArea[iPort].pStart.pVoid));
      osalinux_release_region(pExternalDevice->m_pResources->sqIoPortArea[iPort].pStart, pExternalDevice->m_pResources->sqIoPortArea[iPort].cbLength);
    }

    for (int iMemory = 0; iMemory < pExternalDevice->m_pResources->coMemoryAreas; iMemory++) {
      OSA_DPRINTF(("OsaLinux: Unmapping memory region 0x%p\n", pExternalDevice->m_pResources->sqMemoryRegArea[iMemory].pStart.pVoid));
      osalinux_iounmap(pExternalDevice->m_pResources->sqMemoryRegArea[iMemory].pStart);
    }

    delete pExternalDevice->m_pResources;
    pExternalDevice->m_pResources = NULL;


  } else {
    OSA_DPRINTF(("OsaLinux: No resources to return\n"));
  }
}


/******************************************************************************
******************************************************************************/
int OsaWaitForSingleEvent(OsaEventImpl *pEvent, OsaUint32 nWaitMaxMs)
{
  int nRemainingMs;
  int rc = 0;

  if (!pEvent->m_nEventSet) {
    // Yes, I have thought of the race condition that occurs after we have
    // tested the the m_nEventSet between a SetEvent call and the osalinux_wait_event_timeout
    // call below, however, osalinux_wait_event_timeout tests the m_nEventSet again in
    // a race safe way so that we don't miss the SetEvent call and sleep forever
    rc =  osalinux_wait_event_timeout(&pEvent->m_nEventSet, &pEvent->m_pWaitQHead, nWaitMaxMs);
  }

  return rc;
}

/******************************************************************************
******************************************************************************/
int OsaWaitForSemaphore(OsaSemaphoreImpl *pSemaphore)
{
  OsaSemaphoreImpl_WaitForSemaphore(&pSemaphore->m_Semaphore);
  return 0;
}


/******************************************************************************
******************************************************************************/
OsaResultE OsaGetKernelRefForEventHandle(OsaVoidP hEvent, OsaUserEventP *ppUserEvent)
{
  //OSA_DPRINTF(("OsaLinux: WARNING: OsaGetKernelRefForEventHandle Not implemented\n"));
  // Although we have a device representing the event object living in OSA the user lib
  // actually keps track of the pointer to this device as an opaque pointer as well
  // as the OS handle to the device. In this case we get the opaque pointer passed
  // to us so we just need to cast it.
  *ppUserEvent = (OsaUserEventP) hEvent;

  OSA_DPRINTF(("OsaLinux: OsaGetKernelRefForEventHandle: Event Pointer = %p\n", *ppUserEvent));

  return OSA_RES_SUCCESS;
}

/******************************************************************************
******************************************************************************/
OsaResultE OsaCloseKernelRefForEventHandle(OsaUserEventP ppUserEvent)
{
  return OSA_RES_SUCCESS;
}

/******************************************************************************
******************************************************************************/
OsaBool32 OsaSetUserEvent(OsaUserEventP pUserEvent)
{
  OsaLinuxUserEventDevice *pEvent = reinterpret_cast<OsaLinuxUserEventDevice *>(pUserEvent);

  // Don't forget the user event must be "auto reset" whereas kernel events
  // are "manual reset"
  // Auto-reset is not possible so we have changed the EventQueue to always
  // reset the pointer when extracting from and empty fifo
  if (pEvent) {
    //OSA_DPRINTF(("OsaLinux: OsaSetUserEvent: Event Pointer = %p\n", pEvent));
    if (pEvent->m_pClassValidityObject != (void *) &OsaLinuxUserEventDevice::ms_classValidityObject) {
      OSA_DPRINTF(("OsaSetUserEvent: Not a valid event pointer\n"));
      return OSA_TRUE;
    } else {
      //OSA_DPRINTF(("OsaSetUserEvent: Valid pointer to an event\n"));
      pEvent->SetEvent();
    }
  }

  return OSA_FALSE;
}


/******************************************************************************
******************************************************************************/
void OsaClearUserEvent(OsaUserEventP pUserEvent)
{
  OsaLinuxUserEventDevice *pEvent = reinterpret_cast<OsaLinuxUserEventDevice *>(pUserEvent);

  // Don't forget the user event must be "auto reset" whereas kernel events
  // are "manual reset"
  if (pEvent) {
    //OSA_DPRINTF(("OsaLinux: OsaClearUserEvent: Event Pointer = %p\n", pEvent));
    if (pEvent->m_pClassValidityObject != (void *) &OsaLinuxUserEventDevice::ms_classValidityObject) {
      OSA_DPRINTF(("OsaClearUserEvent: Not a valid event pointer\n"));
    } else {
      //OSA_DPRINTF(("OsaClearUserEvent: Valid pointer to an event\n"));
      pEvent->ClearEvent();
    }
  }
}


/******************************************************************************
******************************************************************************/
bool OsaDmaAdapterCheck(IN OsaExternalDeviceRep pBoard)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}


/******************************************************************************
******************************************************************************/
bool OsaDmaInitAdapter(IN OsaExternalDeviceRep pBoard, OsaULong ulMaxLength, OsaULong ulMaxFragments)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  pExternalDevice->m_DmaMaxMapRegisters = ulMaxFragments;
  return true;
}


/******************************************************************************
******************************************************************************/
bool OsaDmaUnInitAdapter(IN OsaExternalDeviceRep pBoard)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}

/******************************************************************************
******************************************************************************/
OsaULong OsaGetDmaMaxMapRegisters(IN OsaExternalDeviceRep pBoard)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return (OsaULong) pExternalDevice->m_DmaMaxMapRegisters;
}

/******************************************************************************
******************************************************************************/
bool OsaDmaAllocateAdapter(
  IN OsaExternalDeviceRep pBoard,
  IN void * Context
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}

/******************************************************************************
******************************************************************************/
bool OsaDmaFreeAdapter(
  IN OsaExternalDeviceRep pBoard,
  IN void * Context
)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  return true;
}

/******************************************************************************
******************************************************************************/
bool OsaDmaAllocateCommonBuffer(IN OsaExternalDeviceRep pBoard, IN OUT POSA_MD pOmd)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  pOmd->pLog = osalinux_alloc_dma_common_buf(pExternalDevice->m_pPciDev, pOmd->cbSize, (OsaUint64 *) &pOmd->pPhy);
  if (!pOmd->pLog) {
    pOmd->pPhy = 0;
  }

  pOmd->pMdl     = (void *) NULL;  // If not used set to NULL
  pOmd->pContext = (void *) NULL;
  pOmd->ui64Addr = 0;

  return (pOmd->pLog != (void *) NULL);
}

/******************************************************************************
******************************************************************************/
bool OsaDmaFreeCommonBuffer(IN OsaExternalDeviceRep pBoard, IN OUT POSA_MD pOmd)
{
  OsaLinuxDevice *pExternalDevice = reinterpret_cast<OsaLinuxDevice *>(pBoard);

  if (pOmd->pLog) {
    osalinux_free_dma_common_buf(pExternalDevice->m_pPciDev, pOmd->cbSize, (OsaUint64) pOmd->pLog, (OsaUint64) pOmd->pPhy);
  }

  pOmd->pLog = (void *) NULL;
  pOmd->pPhy = (void *) NULL;
  pOmd->ui64Addr = 0;
  pOmd->pContext = (void *) NULL;

  return true;
}


