/* $Workfile: OsaLinuxEntry.c $   $Revision: 4 $    $Modtime: 4/05/13 2:52p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2013, All Rights Reserved

*******************************************************************************/

/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/
// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places

#include <linux/version.h>
#include <linux/interrupt.h>
#include <linux/poll.h>

#include "Lib/OsaIoctl.h"
#include "OsaLinuxEntry.h"
#include "OsaLinuxCpp.h"

// Headers that cannot be included in C++ code even with macro fixes
// hence we need to provide C functions here for which we give access to the 
// C++ parts using our header (OsaLinuxRtl.h), that is written so that it 
// compiles both in C and C++ (with extern "C" linkage in C++, of course)
#include <linux/proc_fs.h>
#include <asm/uaccess.h> //for 2.4, it has to be here, or loading driver fails


#define OSA_WORK_QUEUE 1 // Using a softIsr work queue
// Prints can have a bad effect on the drive 
#define DBG_WORKQUEUE  0 // 1 = debug checking, 2 = more debug, 3 = all
#if DBG_WORKQUEUE != 0
static int TimerIntrpt = 0;      // Total count of work queue activations

extern OsaInt32 OsaGetKernelStackAvailable(OsaULong *stackAvailable);
#define STACK_CHECK {OsaUint64 stackAvailable; OsaGetKernelStackAvailable(&stackAvailable);}
#define OSA_ISR_DPRINTF(pars)     { STACK_CHECK printk pars; }
#define OSA_ISR_EPRINTF(pars)     { printk pars; }

#endif
#if   DBG_WORKQUEUE == 0 
// No debug prints
#define OSA_ISR_DPRINTF(pars)     
#define OSA_ISR_EPRINTF(pars)     { printk pars; }
#define OSA_ISR1_DPRINTF(pars)     
#define OSA_ISR2_DPRINTF(pars)     
#define OSA_ISR3_DPRINTF(pars) 
    
#elif DBG_WORKQUEUE == 1
// Enable error checking with general debug prints
#define OSA_ISR1_DPRINTF(pars)     printk pars
#define OSA_ISR2_DPRINTF(pars)     
#define OSA_ISR3_DPRINTF(pars)     

#elif DBG_WORKQUEUE == 2
// Enable error checking with general debug prints
#define OSA_ISR1_DPRINTF(pars)     printk pars
#define OSA_ISR2_DPRINTF(pars)     printk pars
#define OSA_ISR3_DPRINTF(pars)     

#elif DBG_WORKQUEUE == 3
// Enable all prints (every ISR)
#define OSA_ISR1_DPRINTF(pars)     printk pars
#define OSA_ISR2_DPRINTF(pars)     printk pars
#define OSA_ISR3_DPRINTF(pars)     printk pars

#endif

#if DBG_WORKQUEUE != 0
  #warning ****
  #warning **** Debug prints enabled in OsaLinxEntry.c!
  #warning ****
#endif

#if OSA_WORK_QUEUE == 1
#include <linux/kernel.h>	    // We're doing kernel work
#include <linux/module.h>	    // Specifically, a module
#include <linux/proc_fs.h>	  // Necessary because we use the proc fs
#include <linux/workqueue.h>	// We scheduale tasks here
#include <linux/sched.h>	    // We need to put ourselves to sleep and wake up later
#include <linux/init.h>		    // For __init and __exit
#include <linux/interrupt.h>	// struct tasklet_struct

#define CODE_AUTHOR "Odin TeleSystems (www.odints.com)"
#define CODE_DESC   "Driver with WorkQueue"
// Used to add work to the queue
int OsaLinuxAddWorkQueue(void *tasklets, void *queueData, void (*func)(OsaULong));
static OsaULong *g_nWorkQueueData;
static int g_nWorkQueueFlowCnt = 0; // check to make sure every ISR is processed

void OsaLinuxIsrWorkQueue(OsaULong queueData);

#endif // Using softIsr Handler

extern int OsaLinuxDriverEntry(void);

static char gs_szOsaProcDirName[] = "osadrivers";
static char gs_szStatusProcFileName[] = "status";

static int          CreateOsaProcDir(void);
static void         RemoveOsaProcDir (void);
static ssize_t      OsaLinuxProcReadCallback(struct file* file, char* buf, size_t count, loff_t *ppos);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
static int          OsaLinuxProcPermsCallback (struct inode* inode, int op, struct nameidata *n);
static int          OsaLinuxProcGetInfoCallback(char* buf, char** start, off_t offs, int len);
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,32)
static int          OsaLinuxProcPermsCallback (struct inode* inode, int op);
#else // 2.6.38
static int          OsaLinuxProcPermsCallback (struct inode* inode, int op, unsigned int arg);
#endif


static long         OsaLinuxProcIoControlCallback(struct file *file, unsigned int cmd, OsaULong arg);
static int          OsaLinuxProcOpenCallback(struct inode *inode, struct file *file);
static int          OsaLinuxProcCloseCallback(struct inode *inode, struct file *file);
static unsigned int OsaLinuxPollCallback(struct file *, struct poll_table_struct *);


//#define OSA_GDB_DEBUG 1  // uncomment this line to use the gdb to perform kernel mode debugging
                           // at the startup of the driver


/******************************************************************************
  Generic /proc/osadrivers/<file> file operations
******************************************************************************/
static struct file_operations gs_sProcFileOperations =
{ 
 .owner = THIS_MODULE,

 .read = OsaLinuxProcReadCallback,
 .poll = OsaLinuxPollCallback,

#ifdef HAVE_UNLOCKED_IOCTL
 .unlocked_ioctl = OsaLinuxProcIoControlCallback,
#else
 .ioctl = OsaLinuxProcIoControlCallback,
#endif

 .open = OsaLinuxProcOpenCallback,
 .release = OsaLinuxProcCloseCallback,
};

/******************************************************************************
  Generic /proc/osadrivers/<file> inode operations
******************************************************************************/
static struct inode_operations gs_sProcInodeOperations =
{
  .permission = OsaLinuxProcPermsCallback
};


// following two are used to replace gs_sOsaProcDirNode and gs_sStatusProcFileNode above 
static  struct proc_dir_entry *pOsaDriversDir = NULL;
static  struct proc_dir_entry *pOsaDriversStatus = NULL; // Don't know if we need this global /HJ

																												 
#ifdef OSA_GDB_DEBUG
int odints_open(struct inode *inode, struct file *filep)
{
  int rc;

  OSA_DPRINTF(("OsaLinux: Creating driver\n"));
  rc = OsaLinuxDriverEntry();

  return rc;
}
struct file_operations odintsfops = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	odints_open,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
static int odints_major = 0;
#endif // OSA_GDB_DEBUG

/******************************************************************************
  Driver main exported entry function.
  Called when driver loads
******************************************************************************/
int init_module(void)         
{
  int rc;
  OSA_DPRINTF(("OsaLinux: Loading Osa based driver\n"));

#if 0 // Not used
  {
    int foo;
    OSA_DPRINTF(("OsaLinux: Module parameter nfnOsaLinuxCsIoControl = 0x%X \n", nfnOsaLinuxCsIoControl));
    foo = get_module_symbol("otxdrv_cs", "OsaLinuxCsIoControl");
    OSA_DPRINTF(("OsaLinux: foo = 0x%X \n", foo));
  }
#endif

  // Register the driver as a "Char Device" driver, use dynamic major number assignment
  CreateOsaProcDir();

#ifdef OSA_GDB_DEBUG 
  // Don't run OsaLinuxDriverEntry() quite yet. Run it when we open the dummy driver (odints2)
  // This will allow us to set breakpoints in the startup code.
  if((odints_major = register_chrdev(0, "odints2", &odintsfops)) < 0) {
    printk("OsaLinux: init_odints() register_chrdev failed, error=%d!\n", rc);
    return odints_major;
  }
  return 0;

#else // Normal operation
  OSA_DPRINTF(("OsaLinux: Created proc entries\n"));
  rc = OsaLinuxDriverEntry();

  return rc;   // Something wrong
#endif
}


/******************************************************************************
  Driver main exported exit function
  Called when driver unloads
******************************************************************************/
void cleanup_module(void)         
{
  OSA_DPRINTF(("OsaLinux: Unloading Otx based driver\n"));

#ifdef OSA_GDB_DEBUG 
  if(odints_major > 0) {
		unregister_chrdev(odints_major, "odints2");
  }
#else // Normal operation
  RemoveOsaProcDir();

OsaLinuxUnloadEntry();

#endif

  return;
}


/******************************************************************************
  Create a new device node entry (file) in dir "/proc/osadrivers"  
******************************************************************************/
struct proc_dir_entry_otx *OsaLinuxCreateProcDevice(const wchar_t *pwcsDeviceName, void *pContext)
{
  // no memory allocation for 'de'
  struct proc_dir_entry *de;

  int cbSize = sizeof(char) * (wcslen(pwcsDeviceName) + 1);
  char *pszName = kmalloc(cbSize, GFP_KERNEL);
  wcsztosz(pszName, pwcsDeviceName);

  OSA_DPRINTF(("OsaLinux: Name of proc entry = %s\n", pszName));

  OSA_DPRINTF(("OsaLinux: Proc registering %s, pExternalDeviceP=0x%x\n", pszName, (int) pContext));

  de = create_proc_entry(pszName, 0666 | S_IFREG, pOsaDriversDir);	
  if (!de) {
    OSA_DPRINTF(("OsaLinux: create_proc_entry failed for device = %s\n", pszName));
    return (struct proc_dir_entry_otx *) de;
  }

  de->size = (size_t) pContext;	

  //32bit: de->proc_iops = &gs_sProcInodeOperations;   
  de->proc_fops = &gs_sProcFileOperations;    

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32) //32bit   
  de->get_info = &OsaLinuxProcGetInfoCallback; 
#endif

  return (struct proc_dir_entry_otx *) de;
} 


/******************************************************************************
  Remove the device node entry (file) in dir "/proc/osadrivers"
******************************************************************************/
void OsaLinuxRemoveProcDevice(struct proc_dir_entry_otx *pProcEntry)
{
  OSA_ASSERT(pProcEntry);
  if (pProcEntry) {
    struct proc_dir_entry *l_pProcEntry;
    l_pProcEntry = (struct proc_dir_entry *) pProcEntry ;
    remove_proc_entry(l_pProcEntry->name, pOsaDriversDir);
    OSA_DPRINTF(("OsaLinux: Unregistering proc entry address %p \n", pProcEntry));

  }
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

/* delay removing of /proc entry until after the inode is closed... */
void remove_proc_dev_tasklet_function( unsigned long data )
{
  struct proc_dir_entry_otx *pProcEntry = (void *)data;
  OsaLinuxRemoveProcDevice(pProcEntry);
}

DECLARE_TASKLET( remove_proc_dev_tasklet, remove_proc_dev_tasklet_function, 0 );

void OsaLinuxRemoveProcDeviceDeferred(struct proc_dir_entry_otx *pProcEntry)
{
  remove_proc_dev_tasklet.data = (unsigned long)pProcEntry;
  tasklet_schedule( &remove_proc_dev_tasklet );
}

#endif

/******************************************************************************
  Create the /proc/osadriver directory
******************************************************************************/
static int CreateOsaProcDir(void)
{
  int err = 0;
  OSA_DPRINTF(("OsaLinux: Registering ProcDirNode\n"));

  OSA_DPRINTF(("OsaLinux: Creating %s \n", gs_szOsaProcDirName));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
  pOsaDriversDir = proc_mkdir(gs_szOsaProcDirName, &proc_root);
#else //2.6.32
  pOsaDriversDir = proc_mkdir(gs_szOsaProcDirName, NULL); //proc_root doesn't work
#endif

  if (pOsaDriversDir) {                  
      pOsaDriversStatus = create_proc_entry(gs_szStatusProcFileName, 0666 | S_IFREG, pOsaDriversDir);
               
      if (!pOsaDriversStatus) {
        OSA_DPRINTF(("OsaLinux: Registering StatusProcFileNode failed\n"));
      }
      else {  
        OSA_DPRINTF(("OsaLinux: pOsaDriversStatus address = %p\n", pOsaDriversStatus));                
				pOsaDriversStatus->proc_iops = &gs_sProcInodeOperations; 
        pOsaDriversStatus->proc_fops = &gs_sProcFileOperations;
        
        #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
        pOsaDriversStatus->get_info = &OsaLinuxProcGetInfoCallback;
        #endif        
			}

  } else {
		OSA_DPRINTF(("OsaLinux: Registering gs_szOsaProcDirName (%s)failed\n", gs_szOsaProcDirName));
  }

  return err;
}

/******************************************************************************
  Remove the /proc/osadriver directory and all devices (files) in it
******************************************************************************/
static void RemoveOsaProcDir(void)
{
  struct proc_dir_entry *pOld = NULL;

  // Remove StatusProcFile
	remove_proc_entry(gs_szStatusProcFileName, pOsaDriversDir);

  // Remove all devices in the directory
  while (pOsaDriversDir->subdir && (pOsaDriversDir->subdir != pOld))     {
		pOld = pOsaDriversDir->subdir;
    if (pOsaDriversDir->subdir->size) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)    
      // A proc entry which has a corresponding OSA device (ExternalDevice)
      // The procentry will be removed when the ExternalDevice is removed
      OsaLinuxRemoveDeviceEntry((void *)(pOsaDriversDir->subdir->size));
#else  // 2.6.9
      // A proc entry which has a corresponding OSA device (ExternalDevice)
      // The procentry will be removed when the ExternalDevice is removed
      OsaLinuxRemoveDeviceEntry((void *)pOsaDriversDir->subdir->size);
#endif
    } else {
      // Not at proc entry corresponding to an OSA device
      // Since we unregister the status node first we should never get here
	    OSA_DPRINTF(("OsaLinux: remove %s (else) \n", pOsaDriversDir->subdir->name)); 
      OsaLinuxRemoveProcDevice((struct proc_dir_entry_otx * )pOsaDriversDir->subdir);
    }
  }
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
  remove_proc_entry(gs_szOsaProcDirName, &proc_root);
#else	//2.6.32
  remove_proc_entry(gs_szOsaProcDirName, NULL); 
#endif

}


/******************************************************************************
  Callback, when reading a proc file node (device), for example with the 'cat'
  command
******************************************************************************/
static ssize_t OsaLinuxProcReadCallback(struct file* file, char* buf, size_t count, loff_t *ppos)
{
  OSA_DPRINTF(("OsaLinux: OsaLinuxProcReadCallback: file:%llx  buf:%llx  count:%x  ppos:%llx\n", (OsaUint64) file, (OsaUint64) buf, (int) count, (OsaUint64) ppos));
  return 0;
}

/******************************************************************************
  Callback
******************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
static int OsaLinuxProcPermsCallback (struct inode* inode, int op, struct nameidata *n)
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2,6,32)
static int          OsaLinuxProcPermsCallback (struct inode* inode, int op)
#else // 2.6.38
static int OsaLinuxProcPermsCallback (struct inode* inode, int op, unsigned int arg)
#endif
{
  OSA_DPRINTF(("OsaLinux: OsaLinuxProcPermsCallback\n"));
  return 0;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)    
/******************************************************************************
  Callback
******************************************************************************/
static int OsaLinuxProcGetInfoCallback(char* buf, char** start, off_t offs, int len) 
{
  int cnt;// = sizeof(conf_hdr) - 1;
  OSA_DPRINTF(("OsaLinux: OsaLinuxProcGetInfoCallback\n"));
  strcpy (buf, "Test");
  cnt = strlen (buf);
  return cnt;
}
#endif

/******************************************************************************
******************************************************************************/
int OsaLinuxProcOpenCallback(struct inode *inode, struct file *pFile)
{
  struct proc_dir_entry *pde = PDE(pFile->f_dentry->d_inode);

  OSA_ASSERT(pde);
  if (pde) {
    int rc = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)    
//32bit:    rc = OsaLinuxOpenDeviceEntry(pFile, (void *)((int)pde->size)); 
    rc = OsaLinuxOpenDeviceEntry(pFile, (void *)(pde->size)); 
#else  // 2.6.9
    rc = OsaLinuxOpenDeviceEntry(pFile, (void *)pde->size); 
#endif

    return rc;

  }
  return -1;
}


/******************************************************************************
******************************************************************************/
int OsaLinuxProcCloseCallback(struct inode *inode, struct file *pFile)
{
  struct proc_dir_entry *pde = PDE(inode);

  OSA_ASSERT(pde);
  if (pde) {
    int rc = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)    
//32bit:    rc = OsaLinuxCloseDeviceEntry(pFile, (void *)((int)pde->size)); 
    rc = OsaLinuxCloseDeviceEntry(pFile, (void *)(pde->size)); 
#else   //2.6.9
    rc = OsaLinuxCloseDeviceEntry(pFile, (void *)pde->size); 
#endif

    return rc;
  }
  return -1;
}


/******************************************************************************
  Callback, when user ioctl is called for a device (file) in 
  "/proc/osadrivers/"
******************************************************************************/
//32bit: int OsaLinuxProcIoControlCallback(struct inode *inode, struct file *pFile,
//                unsigned int cmd, unsigned long arg)
long OsaLinuxProcIoControlCallback(struct file *pFile,
				   unsigned int cmd, OsaULong arg)
{
  //32bit: struct proc_dir_entry *pde = PDE(inode);
  struct proc_dir_entry *pde = PDE(pFile->f_dentry->d_inode);

  OsaLinuxIoCtlHdr sHdr;
  void *pSystemBuffer = NULL;
  int cbSystemBufferSize;

// Note: VERIFY_WRITE is a superset of VERIFY_READ
//  if (access_ok(VERIFY_WRITE, (void *) arg, sizeof(OsaLinuxIoCtlHdr)) {

  //OSA_DPRINTF(("OsaLinux: In OsaLinuxProcIoControlCallback, arg = %lx\n", arg));

  // First get the Linux specific header so we can get to the generic parameters
  if (arg) {
    if (copy_from_user(&sHdr, (void *) arg, sizeof(OsaLinuxIoCtlHdr))) { 
      OSA_EPRINTF(("OsaLinux: Unexpected copy_from_user() error (sHdr)\n"));
      return -1;
    }
  } else {
    // Lets make it possible for us to handle NULL arguments, although NULL arg this is abnormal
    memset(&sHdr, 0, sizeof(OsaLinuxIoCtlHdr));
  }

  //OSA_DPRINTF(("OsaLinux: sHdr.pOutBuf = %p\n", sHdr.pOutBuf));

  // Allocate space for the input and output buffers

  cbSystemBufferSize = OSA_MAX(sHdr.cbInBufSize, sHdr.cbOutBufSize); 
  if (cbSystemBufferSize) {
    pSystemBuffer = kmalloc(cbSystemBufferSize, GFP_KERNEL);
    if (!pSystemBuffer) {
      OSA_EPRINTF(("OsaLinux: Unexpected kmalloc() error\n"));
      return -1;
    }
  } else {
    pSystemBuffer = NULL;
  }

  if (sHdr.cbInBufSize) {
    if (copy_from_user(pSystemBuffer, sHdr.pInBuf, sHdr.cbInBufSize)) { 
      OSA_EPRINTF(("OsaLinux: Unexpected copy_from_user() error (sHdr.pInBuf)\n"));
      kfree(pSystemBuffer);
      return -1;
    }
#if 0
/*
    OSA_DPRINTF(("OsaLinux: received system buffer %d bytes in %d out\n", sHdr.cbInBufSize, sHdr.cbOutBufSize));
    {
      int i;
      for (i = 0; i < sHdr.cbInBufSize; i++) {
        OSA_DPRINTF(("%.2x ", ((OsaUint8 *)pSystemBuffer)[i]));
      }
      OSA_DPRINTF(("\n"));
    }
*/
#endif
  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)    
  //32bit: if (OsaLinuxIoDeviceControlEntry(pFile, (void *)((int)pde->size), cmd, &sHdr, pSystemBuffer)) {
  if (OsaLinuxIoDeviceControlEntry(pFile, (void *)(pde->size), cmd, &sHdr, pSystemBuffer)) {
#else  //2.6.9
  if (OsaLinuxIoDeviceControlEntry(pFile, (void *)pde->size, cmd, &sHdr, pSystemBuffer)) {
#endif
    OSA_EPRINTF(("OsaLinux: OsaLinuxIoDeviceControlEntry failed\n"));
    //32bit: if (copy_to_user((void *) arg, &sHdr, sizeof(OsaLinuxIoCtlHdr))) { 
    if (copy_to_user((void *) arg, &sHdr, (OsaULong) sizeof(OsaLinuxIoCtlHdr))) { 
      OSA_EPRINTF(("OsaLinux: Unexpected copy_to_user() error (sHdr)\n"));
      return -1;
    }
    kfree(pSystemBuffer);
    return -1;
  }

  if (sHdr.cbReturned) {
    OSA_ASSERT(sHdr.cbReturned <= cbSystemBufferSize);
    //OSA_DPRINTF(("OsaLinux: Returning %d bytes in sHdr.pOutBuf\n", sHdr.cbReturned));
    if (sHdr.cbReturned <= cbSystemBufferSize) {
      //32bit? if (copy_to_user(sHdr.pOutBuf, pSystemBuffer, sHdr.cbReturned) != 0) {
      if (copy_to_user(sHdr.pOutBuf, pSystemBuffer, (OsaULong) sHdr.cbReturned) != 0) {
        OSA_EPRINTF(("OsaLinux: Unexpected copy_to_user() error (pSystemBuffer)\n"));
        kfree(pSystemBuffer);
        return -1;
      } else {
        //OSA_DPRINTF(("OsaLinux: Copied %d bytes to user mode from pSystemBuffer\n", sHdr.cbReturned)); 
        //32bit: if (copy_to_user((void *) arg, &sHdr, sizeof(OsaLinuxIoCtlHdr))) { 
        if (copy_to_user((void *) arg, &sHdr, (OsaULong) sizeof(OsaLinuxIoCtlHdr))) { 
          OSA_EPRINTF(("OsaLinux: Unexpected copy_to_user() error (sHdr)\n"));
          return -1;
        }
        //OSA_DPRINTF(("OsaLinux: Copied header, %d bytes to user mode\n", sizeof(OsaLinuxIoCtlHdr))); 
        //OSA_DPRINTF(("OsaLinux: <1>Copied Hello to user mode\n"));
      }
    } else {
      return -1;
    }
  }

  if (pSystemBuffer) {
    kfree(pSystemBuffer);
  }
  return 0;
}


/******************************************************************************
  Callback, when user calls poll to ask (and possibly wait) if the device
  is ready for action. i.e. if the device will return or accept data, and do
  so without blocking, if read or written to.
  "/proc/osadrivers/"
******************************************************************************/
static unsigned int OsaLinuxPollCallback(struct file *pFile, struct poll_table_struct *pPollTable)
{
  struct proc_dir_entry *pde = PDE(pFile->f_dentry->d_inode);

  OSA_ASSERT(pde);

  if (pde) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)    
    //32bit: return OsaLinuxPollDeviceEntry(pFile, (void *)((int)pde->size), pPollTable);
    return OsaLinuxPollDeviceEntry(pFile, (void *)(pde->size), pPollTable) ? (POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM) : 0;
#else  //2.6.9
    //32bit: return OsaLinuxPollDeviceEntry(pFile, (void *)pde->size, pPollTable);
    return OsaLinuxPollDeviceEntry(pFile, (void *)pde->size, pPollTable) ? (POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM) : 0;
#endif
  }

  return -1;
}


/******************************************************************************
** This function is called when the kernel has time to allow a task 
** from the work queue to be executed.
**
** Work removed from work queue
******************************************************************************/
#if OSA_WORK_QUEUE == 1
//32bit: void OsaLinuxIsrWorkQueue(unsigned long queueData)
void OsaLinuxIsrWorkQueue(OsaULong queueData)
{

  g_nWorkQueueFlowCnt--; // This value should always be 0

#if DBG_WORKQUEUE >= 1
  // Increment the debug counter
	TimerIntrpt++;
  if(g_nWorkQueueFlowCnt >= 1) {
    // The workqueue has been given more work then it has 
    // been able to process.
    OSA_ISR_DPRINTF(("OsaLinuxIsrWorkQueue: SoftIsr running slow count=%d\n",
                    g_nWorkQueueFlowCnt));
  }

  OSA_ISR3_DPRINTF(("OsaLinuxIsrWorkQueue: count %d\n", TimerIntrpt));
  //32bit: OSA_ISR3_DPRINTF(("OsaLinuxIsrWorkQueue: queueData 0x%x\n", 
  //32bit:                 (unsigned int) queueData));
  OSA_ISR3_DPRINTF(("OsaLinuxIsrWorkQueue: queueData 0x%lx\n", 
                   (OsaULong) queueData));
#endif

#if DBG_WORKQUEUE >= 2
  // Dispaly this count every 1000 times just to show it is working
  if(!(TimerIntrpt%1000)) {
    OSA_ISR_DPRINTF(("OsaLinuxIsrWorkQueue: %d\n", TimerIntrpt));
  }
#endif

  // Finish the ISR's work. This function is in OsaLinux.cpp used
  // to interface to the interrupt object.
  OsaLinuxWorkQueueFunction((void *) queueData);

  return;
}

/******************************************************************************
** Add work to the work queue
******************************************************************************/
int OsaLinuxAddWorkQueue(void *tasklets,      // Workqueue data stucture
                         void *queueData,     // Object pointer
                         //32bit: void (*func)(unsigned long))      // function for workqueue
                         void (*func)(OsaULong))      // function for workqueue
{
  int queueError = 0;
  struct tasklet_struct *ourTasklet;
  //32bit: unsigned long *nWorkQueueData;
  OsaULong *nWorkQueueData;

#if DBG_WORKQUEUE >= 1
  //32bit: static unsigned long tstPtr = 2;
  //32bit: static unsigned long tstFuncPtr = 2;
  static OsaULong tstPtr = 2;
  static OsaULong tstFuncPtr = 2;
#endif

  ourTasklet = *(struct tasklet_struct **) tasklets;
  if(ourTasklet == NULL) {
	  ourTasklet = kmalloc(sizeof(struct tasklet_struct), GFP_ATOMIC);
    if(ourTasklet == NULL) {
      OSA_ISR_EPRINTF(("OsaLinuxAddWorkQueue: kmalloc returned NULL pointer\n"));

      return -1;
    }
	  *(struct tasklet_struct **) tasklets = ourTasklet;

#if 0 //32bit
    nWorkQueueData = (unsigned long *) queueData;
    OSA_ISR_EPRINTF(("OsaLinuxAddWorkQueue: kmalloc ourTasklet is set to 0x%p, *queueData 0x%X, func (0x%p)\n", 
      ourTasklet, 
      (unsigned int) *nWorkQueueData,
      func));
#endif

    nWorkQueueData = (OsaULong *) queueData;
    OSA_ISR_EPRINTF(("OsaLinuxAddWorkQueue: kmalloc ourTasklet is set to 0x%p, *queueData 0x%lx, func (0x%p)\n", 
      ourTasklet, 
      (OsaULong) *nWorkQueueData,
      func));

  }

  g_nWorkQueueFlowCnt++;

  //32bit g_nWorkQueueData = (unsigned long *) queueData;
  g_nWorkQueueData = (OsaULong *) queueData;

#if DBG_WORKQUEUE >= 1

#if 0 //32bit
  if(tstPtr != (unsigned long) ourTasklet) {
    tstPtr = (unsigned long) ourTasklet;
    OSA_ISR_DPRINTF(("WORKQUE SoftIsr tstPtr-2 = 0x%x\n", (unsigned int) tstPtr));
  }

  if(tstFuncPtr != (unsigned long) func) {
    tstFuncPtr = (unsigned long) func;
//    OSA_ISR_DPRINTF(("WORKQUE SoftIsr func = 0x%x\n", (unsigned int) tstFuncPtr));
  }
#endif

  if(tstPtr != (OsaULong) ourTasklet) {
    tstPtr = (OsaULong) ourTasklet;
    OSA_ISR_DPRINTF(("WORKQUE SoftIsr tstPtr-2 = 0x%lx\n", (OsaULong) tstPtr));
  }

  if(tstFuncPtr != (OsaULong) func) {
    tstFuncPtr = (OsaULong) func;
//    OSA_ISR_DPRINTF(("WORKQUE SoftIsr func = 0x%x\n", (unsigned int) tstFuncPtr));
  }
#endif

  //OSA_ISR3_DPRINTF(("queueData 0x%x\n", (unsigned int) queueData));
  //32bit: OSA_ISR3_DPRINTF(("g_nWorkQueueData 0x%x\n", (unsigned int) *g_nWorkQueueData));
  OSA_ISR3_DPRINTF(("g_nWorkQueueData 0x%lx\n", (OsaULong) *g_nWorkQueueData));

  // If the process is not valid then do not put it on the queue.
  if(pOsaDriversStatus != NULL) {
    if(g_nWorkQueueFlowCnt >= 2) {
      // Soft ISR is not being processed or is slow to process
      OSA_ISR_DPRINTF(("OsaLinuxAddWorkQueue: SoftIsr Queue count is 0x%x\n", 
                       //32bit: (unsigned int) g_nWorkQueueFlowCnt));
                       (int) g_nWorkQueueFlowCnt));
 }

    //32bit: nWorkQueueData = (unsigned long *) queueData;
    nWorkQueueData = (OsaULong *) queueData;
    tasklet_init(ourTasklet, func, *nWorkQueueData);
    if(ourTasklet->func != func) {
      OSA_ISR_EPRINTF(("OsaLinuxAddWorkQueue: ourTasklet(0x%p) ourTasklet->func (0x%p) != func (0x%p)\n", 
        ourTasklet, 
        ourTasklet->func, 
        func));
      ourTasklet->func = func;
    }
    if(ourTasklet->data != *nWorkQueueData) {
#if 0 //32bit
      OSA_ISR_EPRINTF(("OsaLinuxAddWorkQueue: ourTasklet(0x%p) ourTasklet->data (0x%X) != *nWorkQueueData (0x%X)\n", 
        ourTasklet, 
        (unsigned int) ourTasklet->data, 
        (unsigned int) *nWorkQueueData));
#endif

      OSA_ISR_EPRINTF(("OsaLinuxAddWorkQueue: ourTasklet(0x%p) ourTasklet->data (0x%lx) != *nWorkQueueData (0x%lx)\n", 
        ourTasklet, 
        (OsaULong) ourTasklet->data, 
        (OsaULong) *nWorkQueueData));

      ourTasklet->data = *nWorkQueueData;
    }

    tasklet_schedule(ourTasklet);
  }
  else {
    queueError = -1;
    OSA_ISR2_DPRINTF(("pOsaDriversStatus is NULL, not initialized\n"));
  }

  return(queueError);
}

/* 
** Some work queue related functions are just available to GPL licensed Modules?
** Need to verify!
*/
MODULE_LICENSE("GPL");

MODULE_AUTHOR(CODE_AUTHOR);
MODULE_DESCRIPTION(CODE_DESC);

#endif




