/* $Workfile: OsaMemDbg.cpp $   $Revision: 2 $    $Modtime: 4/01/13 4:17p $  */
/*******************************************************************************

  Copyright (c) Odin TeleSystems Inc., 1998-2013, All Rights Reserved.

  OTX Telecom Adapter Driver

  @doc internal

  @module

*******************************************************************************/

#include "OsaMemDbg.h"
#include "OsaRtl.h"

#if DBG

#define EXTRA_DPRINTF 0

static OsaULong gs_cbTotalMemory = 0;
static OsaUint32 gs_nSequence = 0;
static OsaMemoryHeader *gs_pLast = NULL;

OsaULong OsaMemoryGetCurrentUsage()
{
  return gs_cbTotalMemory;
}

void OsaDumpMemoryHeader(char *s, void *pUser)
{
  OsaMemoryHeader *pHeader;

  pHeader = (OsaMemoryHeader *) ((char *) pUser - (sizeof(OsaMemoryHeader)));
#if DBG
  pHeader = (OsaMemoryHeader *) ((char *) pHeader - OSA_NO_MANS_LAND_CB);
#endif
  OSA_DPRINTF(("%s: %s(%d) : Block %zu bytes (p=%p, seq=%d)\n", s, pHeader->m_pszFileName, pHeader->m_nLine, pHeader->m_cbUserBlock, pUser, pHeader->m_nSequence));
}


// Print the list of blocks in use
void OsaPrintCurrentlyUsed()
{
  OsaMemoryHeader *pHeader = gs_pLast;
                  
  while (pHeader) {
    OsaUint8 *pUser = ((OsaUint8 *)pHeader) + sizeof(OsaMemoryHeader) + OSA_NO_MANS_LAND_CB;
    OSA_DPRINTF(("Osa: %s(%d) : Block %zu bytes (p=%p, seq=%d)\n", pHeader->m_pszFileName, pHeader->m_nLine, pHeader->m_cbUserBlock, pUser, pHeader->m_nSequence));
    pHeader = pHeader->m_pPrev;
  }

}


// The pointer passed in should point to the base of newly allocated memory.
// This function will return the pointer to the user data area, but first
// it will add a header and a "No mans land" region before and after
// the user data area. The "No mans land" regions are used to detect out
// of bounds accesses. Detection is done when a block is deleted.
// The header contains debug and assertion info.
void *OsaMakeMemoryHeaderTail(void *pBase, size_t cbUserBlock, char *pszFileName, int nLine)
{
  // Make sure to only use byte pointers for pointer arithmetic since we are dealing with byte addresses (raw memory)

  OsaMemoryHeader *pHeader        = (OsaMemoryHeader *) pBase;
  OsaUint8 *pNoMansLandHeader  = ((OsaUint8 *) pBase) + sizeof(OsaMemoryHeader);
  OsaUint8 *pUser              = pNoMansLandHeader + OSA_NO_MANS_LAND_CB;
  OsaUint8 *pNoMansLandTail    = pUser + cbUserBlock;

  gs_cbTotalMemory += cbUserBlock;

  pHeader->m_cbUserBlock    = cbUserBlock;
  pHeader->m_nSequence      = gs_nSequence++;
  pHeader->m_bDeleted       = OSA_FALSE;
  pHeader->m_nLine          = nLine;
  pHeader->m_pszFileName    = pszFileName;

  pHeader->m_pPrev          = gs_pLast;
  pHeader->m_pNext          = NULL;
  if (gs_pLast) {
    gs_pLast->m_pNext       = pHeader;
  }
  gs_pLast                  = pHeader;

  memset(pNoMansLandHeader, OSA_NO_MANS_LAND_VAL, OSA_NO_MANS_LAND_CB);
  memset(pNoMansLandTail, OSA_NO_MANS_LAND_VAL, OSA_NO_MANS_LAND_CB);
  memset(pUser, OSA_NEW_BLOCKS_VAL, cbUserBlock);

#if EXTRA_DPRINTF
  OSA_DPRINTF(("Osa: Allocated %zu bytes (p=%p, seq=%d), usage=%d\n", cbUserBlock, pUser, gs_nSequence - 1, gs_cbTotalMemory));
#endif

  return pUser;
} 

// When deleting a memory block this function must be called to unlink the header before the memory is deleted
void *OsaUndoMemoryHeaderTail(void *pUser)
{
  // Make sure to only use byte pointers for pointer arithmetic since we are dealing with byte addresses (raw memory)
  void *pBase = OsaIsMemoryValid(pUser);
  if (pBase) {
    OsaMemoryHeader *pHeader = (OsaMemoryHeader *) pBase;
    gs_cbTotalMemory -= pHeader->m_cbUserBlock;
#if EXTRA_DPRINTF
    OSA_DPRINTF(("Osa: Deleting, (p=%p, seq=%d), size=%zu, memory left in use = %lu bytes\n", pUser, pHeader->m_nSequence, pHeader->m_cbUserBlock, gs_cbTotalMemory)); 
#endif
    memset(pUser, OSA_FREED_BLOCKS_VAL, pHeader->m_cbUserBlock);

    pHeader->m_bDeleted = OSA_TRUE;
    static OsaUint32 nBreakAt = 0xFFFFFFFF;  // High number to prevent break
    if (pHeader->m_nSequence == nBreakAt) {
      // Set the nBreakAt with the debugger to break at a certain deallocation moment
      OSA_BREAKPOINT();
    }

    // Unlink this block from the list of allocated memory

    // First, if a previously allocated block exists, make it point to the closest block more recently allocated, if any
    //
    if (pHeader->m_pPrev) {
      pHeader->m_pPrev->m_pNext  = pHeader->m_pNext;  // Could be NULL, which is okay
    }

    // Second, if a more recently allocated block exists, make it point to closest, previously allocated block, if any
    //
    if (pHeader->m_pNext) {
      pHeader->m_pNext->m_pPrev  = pHeader->m_pPrev;  // Could be NULL, which is okay
    }

    // Third, Update the head/root pointer in case this block was the most recently allocated
    if (gs_pLast == pHeader) {
      gs_pLast = pHeader->m_pPrev;
    }
  }

  return pBase;
} 


// This function is used to assert that the newly allocated and deleted user data
// area and the "No mans land" regions contain the correct constant value.
OsaBool32 OsaIsMemoryConstByte(void *pMemory, size_t cbMemory, OsaUint8 nConstValue)
{
  OsaUint8 *pStart = (OsaUint8 *) pMemory;
  OsaUint8 *pEnd   = pStart + cbMemory;

  // A bogus address will most likely cause a fatal error here...
  // But things like that should get caught anyway.
  while (pStart != pEnd) if (*pStart++ != nConstValue) return OSA_FALSE;
  return OSA_TRUE;
}


// Validates the "No mans land areas", and adjusts the pointer to one that
// points to the actual beginning of allocated memory, but only if the validations
// succeeds. If the validations fail we don't want to delete the memory. It's better
// to get a memory leak and and a breakpoint during debugging than a system crash.
void *OsaIsMemoryValid(void *pUserVoidPtr)
{
  OsaUint8 *pUser = (OsaUint8 *) pUserVoidPtr;
  OsaUint8 *pBase = (pUser - sizeof(OsaMemoryHeader) - OSA_NO_MANS_LAND_CB);
  OsaMemoryHeader *pHeader = (OsaMemoryHeader *) pBase;
  if (OsaIsMemoryConstByte(pBase + sizeof(OsaMemoryHeader), OSA_NO_MANS_LAND_CB, OSA_NO_MANS_LAND_VAL)) {
    // NoMansLand in front is okay
    if (pHeader->m_bDeleted) {
      // Error: ALready deleted
      OSA_EPRINTF(("Osa: ERROR: invalid memory block (p=%p, seq=%d). Already deleted\n", pUser, pHeader->m_nSequence));
      if (!OsaIsMemoryConstByte(pUser, pHeader->m_cbUserBlock, OSA_FREED_BLOCKS_VAL)) {
        // Error: Its already deleted, but has been clobbered as well
        OSA_EPRINTF(("Osa: ERROR: invalid memory block (p=%p, seq=%d). Deleted block clobbered\n", pUser, pHeader->m_nSequence));
      }
      OSA_BREAKPOINT();
      return OSA_NULL;  
    } else { 
      // Block appears to be in use
      if (OsaIsMemoryConstByte(pUser + pHeader->m_cbUserBlock, OSA_NO_MANS_LAND_CB, OSA_NO_MANS_LAND_VAL)) {
        // NoMansLand at the back is okay as well
        // Make an additional sanity check
        if (OsaIsMemoryConstByte(pUser, pHeader->m_cbUserBlock, OSA_FREED_BLOCKS_VAL)) {
          // Error: header says the block is in use but it's contents is suspiciously like a deleted block
          OSA_DPRINTF(("Osa: WARNING: memory block appears deleted (p=%p, seq=%d)\n", pUser, pHeader->m_nSequence));
        }
        return pBase;  
      } else {
        // Error: NoMansLand at the back is corrupt
        OSA_EPRINTF(("Osa: ERROR: invalid memory block (p=%p, seq=%d). Back-end out-of-bounds access detected\n", pUser, pHeader->m_nSequence));
        OSA_BREAKPOINT();
        return pBase;   // Sure, the memory has been clobbered but it seems the pointer is still okay, so let's delete it (cross your fingers)
      }
    } 
  } else {
    // Error: NoMansLand at the front is corrupt
    OSA_EPRINTF(("Osa: ERROR: invalid memory block (p=%p). Front-end out-of-bounds access detected\n", pUser));
    OSA_BREAKPOINT();
    return OSA_NULL;      // Hard to say if we have a valid pointer at all so let's not delete it.
  }  
}

#else   // #if DBG

#endif  // #if DBG
