/* $Workfile: OsaRtl.h $   $Revision: 3 $    $Modtime: 4/05/13 4:26p $  */
/*******************************************************************************

  Copyright (c) Odin TeleSystems Inc., 1998-2013, All Rights Reserved.

  OTX Telecom Adapter Driver

  @doc internal

  @module

*******************************************************************************/

#ifndef OSA_RTL_H
#define OSA_RTL_H

#include <Api/OsaDef.h>

// Some of these functions are sometimes implemented directly by the kernel
// but for some configurations they are implemented in OSA
#define OSA_RTL_H_INCLUDING
#if OSA_WIN32
#include "Win32/OsaWin32Rtl.h"
#elif OSA_LINUX
#include "Linux/OsaLinuxRtl.h"
#else
#error "Must define either OSA_LINUX or OSA_WIN32"
#endif
#undef OSA_RTL_H_INCLUDING

// Executions levels. The execution level determines what OSA services
// can be used. A higher level means more restrictions in what OSA functions can
// be called. All levels can be interrupted/preempted by code running on the same level
// or higher. For EXE_LEVEL_CONTROL and EXE_LEVEL_SYSTEM OdOsaCriticalSections can be used
// to protect shared resources. If a resource is used by EXE_LEVEL_DEVICE code it must be
// protected using DeviceLevelSafe callbacks.
typedef enum OsaExeLevel { 
  OSA_EXE_LEVEL_CONTROL = 0, // The least restrictive level, similar to a user thread.
  OSA_EXE_LEVEL_SYSTEM  = 1, // The same priority as the thread scheduler, hence no thread like functionality is available (e.g. "sleep" and "wait" functions).
  OSA_EXE_LEVEL_DEVICE  = 2  // The highest priority, All other activity in the system is blocked, e.g. in the interrupt handler for a device.
} OsaExeLevel;

typedef enum OsaThreadExitCode {
	OSA_THREAD_EXITCODE_INVALID		= 0,	// Invalid exit code
	OSA_THREAD_EXITCODE_EXITED		= 1, // Thread exited
	OSA_THREAD_EXITCODE_STILLACTIVE	= 2  // Thread e
}OsaThreadExitCode;

typedef enum OsaResultE
{
  OSA_RES_SUCCESS             = 0,
  OSA_RES_FAILED              = -1,
  OSA_RES_TIMEOUT             = -2,
  OSA_RES_NOT_IMPL            = -3,
  OSA_RES_NO_MEMORY           = -4,
  OSA_RES_INVALID_PARAMETER   = -5,
  OSA_RES_NOT_SUPPORTED       = -6,
  OSA_RES_INVALID_BUFFER_SIZE = -7,
  OSA_RES_INVALID_REQUEST     = -8,
  OSA_RES_BUSY                = -9
} OsaResultE;


#define OSA_BRD_IOMEM_AREAS_MAX 4    // Max number of board address areas that we can handle (the original value was 2. okay to increase, e.g. to 6, which is PCI max)

#if defined(_M_AMD64)
#define OSA_FIELD_OFFSET(type, field)    ((OsaUint64)(char *)&(((type *)0)->field))
#else
#define OSA_FIELD_OFFSET(type, field)    ((OsaULong)(char *)&(((type *)0)->field))
#endif

#define OSA_MAKEPTR(address) ((void *) (address))

/*
**  Values are 32 bit values layed out as follows:
**
**   3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
**   1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
**  +---+-+-+-----------------------+-------------------------------+
**  |Sev|C|R|     Facility          |               Code            |
**  +---+-+-+-----------------------+-------------------------------+
**
**  where
**
**      Sev - is the severity code
**
**          00 - Success
**          01 - Informational
**          10 - Warning
**          11 - Error
**
**      C - is the Customer code flag
**
**      R - is a reserved bit
**
**      Facility - is the facility code
**
**      Code - is the facility's status code
**
*/

// Used with event log to determine event log error level
#define OSA_INFORMATIONAL_EVENT_LOG   ((OsaUint32)0x48000000L)
#define OSA_WARNING_EVENT_LOG         ((OsaUint32)0x88000000L)
#define OSA_ERROR_EVENT_LOG           ((OsaUint32)0xC8000000L)

#ifdef __cplusplus
extern "C" {
#endif

//struct OsaIrp;

//typedef struct OsaIrpLink {
//  struct OsaIrp *pIrpNext;
//} OsaIrpLink;

                                                                                                                                                                                                                               
typedef struct OsaIrp {
  OsaUint32   nIoCode;
  void        *pInputBuffer;     // Contains the input struct/buffer
  void        *pOutputBuffer;    // Place for the output struct/buffer, NOTE: space most likely shared with input
  OsaUint32   cbInput;           // Size in bytes of the input struct/buffer
  OsaUint32   cbOutputCapacity;  // Max size allowed for the output struct (size of users output buffer)
  OsaUint32   cbOutput;          // Actual size of the output struct/buffer
  OsaResultE  result;            //
#if defined(_M_AMD64)
  PFILE_OBJECT   nOwnerId;       // File object reference (each process gets it's own file object even if the associated device is the same)
#else
  OsaULong    nOwnerId;          // File object reference (each process gets it's own file object even if the associated device is the same)
#endif
//  int coAssocIrps;
//  struct OsaIrp *pMasterIrp;
//  OsaIrpLink  sLink;
} OsaIrp, *OsaIrpP;


//typedef struct OsaIrpQueue {
//  OsaIrpLink sLink;
//} OsaIrpQueue, *OsaIrpQueueP;

OsaUint32   OsaIrpGetIoCode(IN OsaIrpP pIrp);
OsaVoidP    OsaIrpGetInputBuffer(IN OsaIrpP pIrp);
OsaVoidP    OsaIrpGetOutputBuffer(IN OsaIrpP pIrp);
#if defined(_M_AMD64)
OsaUint32    OsaIrpGetOwnerId(IN OsaIrpP pIrp);
#else
OsaULong    OsaIrpGetOwnerId(IN OsaIrpP pIrp);
#endif
OsaUint32   OsaIrpGetInputBufferSize(IN OsaIrpP pIrp);
OsaUint32   OsaIrpGetOutputBufferCapacity(IN OsaIrpP pIrp);
//void        OsaIrpMarkIrpPending(IN OsaIrpP pIrp);
OsaResultE  OsaIrpComplete(IN OsaIrpP pIrp, IN OsaResultE eResult, IN OsaUint32 cbOutput);

//OsaIrpP OsaIrpAllocate();
//OsaIrpP OsaIrpAllocateAssocIrp(OsaIrpP pMasterIrp);
//void    OsaIrpDelete(OsaIrpP pIrp);

//void    OsaIrpQueueAdd(IN OsaIrpQueueP pIrpQueue, IN OsaIrpP pIrp);
//OsaIrpP OsaIrpQueueExtract(IN OsaIrpQueueP pIrpQueue);

#if defined(_M_AMD64)
OsaUint8    OsaInterlockedModifyRegBits8(OsaUint64 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits);
OsaUint8    OsaInterlockedModifyPortBits8(OsaUint64 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits);
#else
OsaUint8    OsaInterlockedModifyRegBits8(OsaULong pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits);
OsaUint8    OsaInterlockedModifyPortBits8(OsaULong pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits);
#endif

void        OsaDelayThreadMicroSecs(OsaInt64 nMicroSecs);
void        OsaStallExecutionMicroSecs(IN OsaUint32 nMicroSecs);

OsaResultE  OsaSetDriverParameters(void);

OsaExeLevel OsaGetCurrentExeLevel(void);

void        OsaQuerySystemTime(OsaDateTime *pDateTime);

OsaUint32   OsaGetSystemTimeTickFrequency(void);

void        OsaSetSystemTimeSource(OsaBool32 bUse1601TypeTimeStamp);

OsaUint32   OsaQueryClockMs(void);

OsaResultE  OsaGetKernelRefForEventHandle(OsaVoidP hEvent, OsaUserEventP *ppUserEvent);
OsaResultE  OsaCloseKernelRefForEventHandle(OsaUserEventP ppUserEvent);

OsaBool32   OsaSetUserEvent(OsaUserEventP pUserEvent);

void        OsaClearUserEvent(OsaUserEventP pUserEvent);

OsaUint64		OsaInt64div32(OsaUint64 divident, OsaUint32 divider);

// generic user memory descriptor block structure
//    we use this structure when locking down/accessing
//    user mode memory (non-kernel memory)
typedef struct {
  OsaVoidP    hMemory;            // generic memory descriptor
  OsaVoidP    pUserBuffer;        // address of unlocked (paged) user buffer
  OsaULong    iUserBufferSize;    // size of user buffer
  OsaVoidP    pLockedBuffer;      // address of locked (non-paged) memory
} OsaUserMemoryDesc;

#if defined(_M_AMD64)
OsaVoidP OsaLockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD, IN OsaUint32 pUserBuffer, IN OsaUint32 uiUserbufferSize);
#else
OsaVoidP OsaLockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD, IN OsaVoidP pUserBuffer, IN OsaUint32 uiUserbufferSize);
#endif


#if defined(_M_AMD64)
OsaVoidP OsaLockUserMemoryUint(IN OUT OsaUserMemoryDesc *pOsaUMD, IN OsaUint64 pUserBuffer, IN OsaUint32 uiUserbufferSize);
#else
OsaVoidP OsaLockUserMemoryUint(IN OUT OsaUserMemoryDesc *pOsaUMD, IN OsaUint32 pUserBuffer, IN OsaUint32 uiUserbufferSize);
#endif


OsaResultE OsaUnlockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD);

// platform type - enabling us to make run-time decisions
typedef enum OsaPlatformE {
  OSA_PLATFORM_UNDEFINED      = 0,
  OSA_PLATFORM_CE_STRONGARM   = 1,
  OSA_PLATFORM_CE_XSCALE      = 2
} OsaPlatformE;

extern OsaPlatformE gOsaPlatform;

#ifdef __cplusplus
}
#endif

#endif  // OSA_RTL_H









